# -*- coding: utf-8 -*-

'''
Functions for processing Japanese text

Latest version can be found at https://github.com/letuananh/chirptext

References:
    MeCab homepage:
        http://taku910.github.io/mecab/

MeCab, デコ, got the joke?
* This script was a part of omwtk

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import re
import os
import logging
import warnings

from . import texttaglib as ttl
from .dekomecab import wakati, parse as _internal_mecab_parse, _register_mecab_loc as set_mecab_bin, _get_mecab_loc as get_mecab_bin


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

DATA_FOLDER = os.path.abspath(os.path.expanduser('./data'))

# reference: https://en.wikipedia.org/wiki/Hiragana_%28Unicode_block%29
# U+304x 		ぁ 	あ 	ぃ 	い 	ぅ 	う 	ぇ 	え 	ぉ 	お 	か 	が 	き 	ぎ 	く
# U+305x 	ぐ 	け 	げ 	こ 	ご 	さ 	ざ 	し 	じ 	す 	ず 	せ 	ぜ 	そ 	ぞ 	た
# U+306x 	だ 	ち 	ぢ 	っ 	つ 	づ 	て 	で 	と 	ど 	な 	に 	ぬ 	ね 	の 	は
# U+307x 	ば 	ぱ 	ひ 	び 	ぴ 	ふ 	ぶ 	ぷ 	へ 	べ 	ぺ 	ほ 	ぼ 	ぽ 	ま 	み
# U+308x 	む 	め 	も 	ゃ 	や 	ゅ 	ゆ 	ょ 	よ 	ら 	り 	る 	れ 	ろ 	ゎ 	わ
# U+309x 	ゐ 	ゑ 	を 	ん 	ゔ 	ゕ 	ゖ 			゙ 	゚ 	゛ 	゜ 	ゝ 	ゞ 	ゟ

HIRAGANA = 'ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼぽまみむめもゃやゅゆょよらりるれろゎわゐゑをんゔゕゖ゙゚゛゜ゝゞゟ'

# reference: https://en.wikipedia.org/wiki/Katakana_%28Unicode_block%29
# U+30Ax 	゠ 	ァ 	ア 	ィ 	イ 	ゥ 	ウ 	ェ 	エ 	ォ 	オ 	カ 	ガ 	キ 	ギ 	ク
# U+30Bx 	グ 	ケ 	ゲ 	コ 	ゴ 	サ 	ザ 	シ 	ジ 	ス 	ズ 	セ 	ゼ 	ソ 	ゾ 	タ
# U+30Cx 	ダ 	チ 	ヂ 	ッ 	ツ 	ヅ 	テ 	デ 	ト 	ド 	ナ 	ニ 	ヌ 	ネ 	ノ 	ハ
# U+30Dx 	バ 	パ 	ヒ 	ビ 	ピ 	フ 	ブ 	プ 	ヘ 	ベ 	ペ 	ホ 	ボ 	ポ 	マ 	ミ
# U+30Ex 	ム 	メ 	モ 	ャ 	ヤ 	ュ 	ユ 	ョ 	ヨ 	ラ 	リ 	ル 	レ 	ロ 	ヮ 	ワ
# U+30Fx 	ヰ 	ヱ 	ヲ 	ン 	ヴ 	ヵ 	ヶ 	ヷ 	ヸ 	ヹ 	ヺ 	・ 	ー 	ヽ 	ヾ 	ヿ
KATAKANA = '゠ァアィイゥウェエォオカガキギクグケゲコゴサザシジスズセゼソゾタダチヂッツヅテデトドナニヌネノハバパヒビピフブプヘベペホボポマミムメモャヤュユョヨラリルレロヮワヰヱヲンヴヵヶヷヸヹヺ・ーヽヾヿ'
KATA2HIRA_TRANS = str.maketrans(KATAKANA[1:87], HIRAGANA[:86])


def getLogger():
    return logging.getLogger(__name__)


def simple_kata2hira(input_str):
    return input_str.translate(KATA2HIRA_TRANS)


def is_kana(text):
    ''' Check if a text if written in kana only (hiragana & katakana)
    if text is empty then return True
    '''
    if text is None:
        raise ValueError("text cannot be None")
    for c in text:
        if c not in HIRAGANA and c not in KATAKANA:
            return False
    return True


try:
    from jaconv import kata2hira
except:
    # if jaconv is not available, use built-in method
    kata2hira = simple_kata2hira


# -------------------------------------------------------------------------------
# Data Structures
# -------------------------------------------------------------------------------

# Reference: http://taku910.github.io/mecab/#parse
# MeCabToken = namedtuple('MeCabToken', 'surface pos sc1 sc2 sc3 inf conj root reading pron'.split())
class MeCabToken(object):
    def __init__(self, surface, pos=None, sc1=None, sc2=None, sc3=None, inf=None, conj=None, root=None, reading=None, pron=None):
        self.surface = surface
        self.pos = pos
        self.sc1 = sc1
        self.sc2 = sc2
        self.sc3 = sc3
        self.inf = inf
        self.conj = conj
        self.root = root
        self.reading = reading
        self.pron = pron
        self.is_eos = (surface == 'EOS' and pos == sc1 == sc2 == sc3 == inf == conj == root == reading == pron == '')

    def __str__(self):
        return '[{sur}({pos}-{s1}/{s2}/{s3}|{ro}|{re}|{pr})]'.format(sur=self.surface, pos=self.pos, s1=self.sc1, s2=self.sc2, s3=self.sc3, ro=self.root, re=self.reading, pr=self.pron)

    def __repr__(self):
        return str(self)

    def reading_hira(self):
        return kata2hira(self.reading)

    def need_ruby(self):
        return self.reading and self.reading != self.surface and self.reading_hira() != self.surface

    def pos3(self):
        ''' Use pos-sc1-sc2 as POS '''
        parts = [self.pos]
        if self.sc1 and self.sc1 != '*':
            parts.append(self.sc1)
            if self.sc2 and self.sc2 != '*':
                parts.append(self.sc2)
        return '-'.join(parts)

    def to_ruby(self):
        ''' Convert one MeCabToken into HTML '''
        if self.need_ruby():
            surface = self.surface
            reading = self.reading_hira()
            return '<ruby><rb>{sur}</rb><rt>{read}</rt></ruby>'.format(sur=surface, read=reading)
        elif self.is_eos:
            return ''
        else:
            return self.surface

    @staticmethod
    def parse(raw):
        parts = re.split('\t|,', raw)
        if len(parts) < 10:
            parts += [''] * (10 - len(parts))
        (surface, pos, sc1, sc2, sc3, inf, conj, root, reading, pron) = parts
        return MeCabToken(surface, pos, sc1, sc2, sc3, inf, conj, root, reading, pron)

    def to_csv(self):
        return '{sur}\t{pos}\t{s1}\t{s2}\t{s3}\t{inf}\t{conj}\t{ro}\t{re}\t{pr}'.format(sur=self.surface, pos=self.pos, s1=self.sc1, s2=self.sc2, s3=self.sc3, inf=self.inf, conj=self.conj, ro=self.root, re=self.reading, pr=self.pron)


class MeCabSent(object):

    def __init__(self, surface, tokens):
        self.surface = surface
        self.tokens = []
        if tokens:
            self.tokens.extend(tokens)

    def to_ruby(self):
        s = []
        for x in self.tokens:
            s.append(x.to_ruby())
        stext = ' '.join(s)
        # clean sentence a bit ...
        stext = stext.replace(' 。', '。').replace('「 ', '「').replace(' 」', '」').replace(' 、 ', '、').replace('（ ', '（').replace(' ）', '）')
        return stext

    def __getitem__(self, value):
        return self.tokens[value]

    def __len__(self):
        return len(self.tokens)

    @property
    def words(self):
        return [x.surface for x in self.tokens if not x.is_eos]

    @property
    def text(self):
        return self.surface if self.surface else str(self)

    def __repr__(self):
        return "MeCabSent({})".format(repr(self.text))

    def __str__(self):
        return ' '.join([x.surface for x in self.tokens if not x.is_eos])

    def to_ttl(self):
        tsent = ttl.Sentence(self.surface)
        tsent.import_tokens(self.words)
        for mtk, tk in zip((tk for tk in self if not tk.is_eos), tsent):
            tk.pos = mtk.pos3()
            if mtk.root and mtk.root != '*':
                tk.lemma = mtk.root
            else:
                tk.lemma = mtk.surface
            tk.new_tag(mtk.reading_hira(), tagtype="reading", source=ttl.Tag.MECAB)
        return tsent

    @staticmethod
    def parse(text, **kwargs):
        ''' Use mecab to parse one sentence '''
        mecab_out = _internal_mecab_parse(text, **kwargs).splitlines()
        tokens = [MeCabToken.parse(x) for x in mecab_out]
        return MeCabSent(text, tokens)


class DekoText(object):

    def __init__(self, name=''):
        self.sents = []
        self.name = name

    def __len__(self):
        return len(self.sents)

    def __getitem__(self, name):
        return self.sents[name]

    def add(self, sentence_text, **kwargs):
        ''' Parse a text string and add it to this doc '''
        sent = MeCabSent.parse(sentence_text, **kwargs)
        self.sents.append(sent)
        return sent

    def __str__(self):
        return "\n".join(["#{}. {}".format(idx + 1, x) for idx, x in enumerate(self)])

    def to_ttl(self, name=None):
        doc = ttl.Document(name=name if name else self.name)
        for sent in self.sents:
            doc.add_sent(sent.to_ttl())
        return doc

    @staticmethod
    def parse(text, splitlines=True, auto_strip=True, **kwargs):
        doc = DekoText()
        if not splitlines:
            # surface is broken right now ...
            tokens = MeCabSent.parse(text, **kwargs)
            doc.sents = tokenize_sent(tokens, text, auto_strip)
        else:
            lines = text.splitlines()
            for line in lines:
                if auto_strip:
                    doc.add(line.strip(), **kwargs)  # auto-parse
                else:
                    doc.add(line, **kwargs)  # auto-parse
        return doc


# -------------------------------------------------------------------------------
# MeCab helper functions
# -------------------------------------------------------------------------------

def _lines2mecab(lines, **kwargs):
    ''' Use mecab to parse many lines '''
    sents = []
    for line in lines:
        sent = MeCabSent.parse(line, **kwargs)
        sents.append(sent)
    return sents


# -------------------------------------------------------------------------------
# Legacy functions
# -------------------------------------------------------------------------------

def txt2mecab(text, **kwargs):
    warnings.warn("txt2mecab() is deprecated and will be removed in near future. Use deko.parse() instead.", DeprecationWarning, stacklevel=2)
    return MeCabSent.parse(text, **kwargs)


def lines2mecab(lines, **kwargs):
    warnings.warn("lines2mecab() is deprecated and will be removed in near future.", DeprecationWarning, stacklevel=2)
    return _lines2mecab(lines, **kwargs)


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------

def tokenize(content, **kwargs):
    ''' Sentence to a list of tokens (string) '''
    # TODO: Check if wakati better?
    # return wakati(content).split(' ')
    return MeCabSent.parse(content, **kwargs).words


# TODO: Need to calculate cfrom, cto to get surfaces
def tokenize_sent(mtokens, raw='', auto_strip=True):
    ''' Tokenize a text to multiple sentences '''
    sents = []
    bucket = []
    cfrom = 0
    cto = 0
    token_cfrom = 0
    logger = getLogger()
    logger.debug("raw text: {}".format(raw))
    logger.debug("tokens: {}".format(mtokens))
    for t in mtokens:
        if t.is_eos:
            continue
        token_cfrom = raw.find(t.surface, cto)
        cto = token_cfrom + len(t.surface)  # also token_cto
        logger.debug("processing token {} <{}:{}>".format(t, token_cfrom, cto))
        bucket.append(t)
        # sentence ending
        if t.pos == '記号' and t.sc1 == '句点':
            sent_text = raw[cfrom:cto]
            getLogger().debug("sent_text = {} | <{}:{}>".format(sent_text, cfrom, cto))
            if auto_strip:
                sent_text = sent_text.strip()
            sents.append(MeCabSent(sent_text, bucket))
            logger.debug("Found a sentence: {}".format(sent_text))
            cfrom = cto
            bucket = []
    if bucket:
        logger.debug("Bucket is not empty: {}".format(bucket))
        sent_text = raw[cfrom:cto]
        logger.debug("remaining text: {} [{}:{}]".format(sent_text, cfrom, cto))
        if auto_strip:
            sent_text = sent_text.strip()
        sents.append(MeCabSent(sent_text, bucket))
    return sents


def analyse(content, splitlines=True, format=None, **kwargs):
    ''' Japanese text > tokenize/txt/html '''
    sents = DekoText.parse(content, splitlines=splitlines, **kwargs)
    doc = []
    final = sents
    # Generate output
    if format == 'html':
        for sent in sents:
            doc.append(sent.to_ruby())
        final = '<br/>\n'.join(doc)
    elif format == 'csv':
        for sent in sents:
            doc.append('\n'.join([x.to_csv() for x in sent]) + '\n')
        final = '\n'.join(doc)
    elif format == 'txt':
        final = '\n'.join([str(x) for x in sents])
    return final


# useful alias
parse = MeCabSent.parse
parse_doc = DekoText.parse



# -*- coding: utf-8 -*-

'''
Chirptext Arsenal - Cache utilities

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import uuid
import logging
import sqlite3
import zlib

from .leutile import FileHelper


# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


class JiCache:
    ''' Cache textual/binary content using an SQLite file (with internal blob or an external folder)
    '''
    def __init__(self, location, blob_location=None, use_internal_blob=True):
        self.location = os.path.abspath(os.path.expanduser(location))
        self.blob_location = os.path.abspath(os.path.expanduser(blob_location)) if blob_location else self.location + '.blob'
        self.use_internal_blob = use_internal_blob
        logger = getLogger()
        logger.info("Cache DB location    : {location}".format(location=self.location))
        if not use_internal_blob:
            logger.info("Blob location        : {blob_loc}".format(blob_loc=self.blob_location))
        else:
            logger.info("Store BLOB internally: {storage_mode}".format(storage_mode=self.use_internal_blob))
        # self.setup()

    def setup(self):
        if not os.path.isfile(self.location) or os.path.getsize(self.location) == 0:
            getLogger().debug("Setting up DB")
            # create dir to store blobs
            if not self.use_internal_blob:
                FileHelper.create_dir(self.blob_location)
            with sqlite3.connect(self.location) as conn:
                c = conn.cursor()
                # Create tables
                c.execute('''CREATE TABLE IF NOT EXISTS cache_entries (key UNIQUE, value);''')
                c.execute("""CREATE INDEX IF NOT EXISTS CACHE_ENTRIES_KEY_INDEX ON cache_entries (key);""")
                c.execute('''CREATE TABLE IF NOT EXISTS blob_entries (key UNIQUE, compressed, blob_data);''')
                c.execute("""CREATE INDEX IF NOT EXISTS BLOB_ENTRIES_KEY_INDEX ON blob_entries (key);""")
                conn.commit()

    def get_conn(self):
        self.setup()
        return sqlite3.connect(self.location)

    def count_entries(self):
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                c.execute("SELECT COUNT(value) FROM cache_entries")
                return c.fetchone()[0]
            except:
                getLogger().exception("Cannot count entries")
                return None

    def retrieve_keys(self):
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                c.execute("SELECT key FROM cache_entries")
                rows = c.fetchall()
                if rows:
                    return [x[0] for x in rows]
            except:
                getLogger().exception("Cannot retrieve keys")
                return None

    def __retrieve(self, key):
        ''' Retrieve file location from cache DB
        '''
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                if key is None:
                    c.execute("SELECT value FROM cache_entries WHERE key IS NULL")
                else:
                    c.execute("SELECT value FROM cache_entries WHERE key = ?", (key,))
                result = c.fetchone()
                if result is None or len(result) != 1:
                    getLogger().info("There's no entry with key={key}".format(key=key))
                    return None
                else:
                    return result[0]
            except:
                getLogger().exception("Cannot retrieve")
                return None

    def has_key(self, key):
        return key in self

    def __contains__(self, key):
        return self.__retrieve(key) is not None

    def __insert(self, key, value):
        '''
        Insert a new key to database
        '''
        if key in self:
            getLogger().warning("Cache entry exists, cannot insert a new entry with key='{key}'".format(key=key))
            return False
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                c.execute("INSERT INTO cache_entries (key, value) VALUES (?,?)", (key, value))
                conn.commit()
                return True
            except Exception as e:
                # NOTE: A cache error can be forgiven, no?
                getLogger().debug("Cache Error: Cannot insert | Detail = %s" % (e,))
                return False

    def __delete(self, key):
        ''' Delete file key from database
        '''
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                c.execute("DELETE FROM cache_entries WHERE key = ?", (key,))
                conn.commit()
            except:
                getLogger().exception("Cannot delete")
                return None

    INTERNAL_BLOB = "JICACHE_INTERNAL_BLOB"

    def __insert_internal_blob(self, key, blob, compressed=True):
        ''' This method will insert blob data to blob table
        '''
        with self.get_conn() as conn:
            conn.isolation_level = None
            c = conn.cursor()
            try:
                compressed_flag = 1 if compressed else 0
                if compressed:
                    blob = zlib.compress(blob)
                c.execute("BEGIN")
                c.execute("INSERT INTO cache_entries (key, value) VALUES (?,?)", (key, JiCache.INTERNAL_BLOB))
                c.execute("INSERT INTO blob_entries (key, compressed, blob_data) VALUES (?,?,?)", (key, compressed_flag, sqlite3.Binary(blob),))
                c.execute("COMMIT")
                return True
            except:
                getLogger().debug("Cannot insert")
                return False

    def __delete_internal_blob(self, key):
        ''' This method will insert blob data to blob table
        '''
        with self.get_conn() as conn:
            conn.isolation_level = None
            try:
                c = conn.cursor()
                c.execute("BEGIN")
                if key is None:
                    c.execute("DELETE FROM cache_entries WHERE key IS NULL")
                    c.execute("DELETE FROM blob_entries WHERE KEY IS NULL")
                else:
                    c.execute("DELETE FROM cache_entries WHERE key = ?", (key,))
                    c.execute("DELETE FROM blob_entries WHERE KEY = ?", (key,))
                c.execute("COMMIT")
            except:
                getLogger().debug("Cannot delete")
                return False
            return True

    def __retrieve_internal_blob(self, key):
        ''' Retrieve file location from cache DB
        '''
        logger = getLogger()
        with self.get_conn() as conn:
            try:
                c = conn.cursor()
                if key is None:
                    c.execute("SELECT compressed, blob_data FROM blob_entries WHERE KEY IS NULL")
                else:
                    c.execute("SELECT compressed, blob_data FROM blob_entries WHERE KEY = ?", (key,))
                result = c.fetchone()
                if not result:
                    logger.debug("There's no blob entry with key={key}".format(key=key))
                    logger.debug("result = {res}".format(res=result))
                    return None
                else:
                    compressed, blob_data = result
                    logger.debug("retrieving internal BLOB (key={key} | len={ln} | compressed={c})".format(key=key, ln=len(blob_data), c=compressed))
                    return blob_data if not compressed else zlib.decompress(blob_data)
            except:
                getLogger().exception("Cannot retrieve internal blob (key={})".format(key))
                return None
            return True

    def insert_blob(self, key, blob):
        if self.__retrieve(key) is not None:
            getLogger().warning("Key exists, cannot insert entry with key='{key}'".format(key=key))
            return False
        elif self.use_internal_blob:
            # insert to database
            getLogger().debug("Insert BLOB internally")
            # TODO: should the compressed field be in the cache_entries table instead?
            return self.__insert_internal_blob(key, blob, compressed=True)
        else:
            getLogger().debug("Inserting a new blob with key = {key}".format(key=key))
            # Allocate a new cache file
            blob_key = str(uuid.uuid1())
            blob_file = os.path.join(self.blob_location, blob_key)
            while os.path.exists(blob_file):
                blob_key = str(uuid.uuid1())
                blob_file = os.path.join(self.blob_location, blob_key)
                # try to write BLOB content to file
            FileHelper.save(blob_file, blob)
            if not self.__insert(key, blob_key):
                # Cannot insert key, delete the file
                os.unlink(blob_file)
                return False
            else:
                return True

    def retrieve_blob(self, key, encoding=None):
        ''' Retrieve blob in binary format (or string format if encoding is provided) '''
        blob_key = self.__retrieve(key)
        if blob_key is None:
            return None
        if not blob_key:
            raise Exception("Invalid blob_key")
        elif blob_key == JiCache.INTERNAL_BLOB:
            blob_data = self.__retrieve_internal_blob(key)
            return blob_data if not encoding else blob_data.decode(encoding)
        else:
            getLogger().debug("Key[{key}] -> [{blob_key}]".format(key=key, blob_key=blob_key))
            blob_file = os.path.join(self.blob_location, blob_key)
            return FileHelper.read(blob_file)

    def delete_blob(self, key):
        blob_key = self.__retrieve(key)
        logger = getLogger()
        if blob_key is None:
            logger.debug("Nothing to delete for key = {}".format(key))
        elif blob_key == JiCache.INTERNAL_BLOB:
            # blah
            logger.debug("deleting internal BLOB")
            self.__delete_internal_blob(key)
        else:
            logger.debug("deleting BLOB from file")
            logger.debug("Blob loc: {}".format(self.blob_location))
            logger.debug("Blob key: {}".format(blob_key))
            blob_file = os.path.join(self.blob_location, blob_key)
            os.unlink(blob_file)
            self.__delete(key)

    def insert_string(self, key, a_string, encoding='utf-8'):
        return self.insert_blob(key, a_string.encode(encoding))

    def retrieve_string(self, key, encoding='utf-8'):
        return self.retrieve_blob(key, encoding=encoding)

    def insert_file(self, key, file_path):
        with open(file_path, 'rb') as input_file:
            byte_data = input_file.read()
            return self.insert_blob(key, byte_data)



# -*- coding: utf-8 -*-

'''
Tools for processing Chinese

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import logging
from collections import defaultdict as dd

from . import chio
from .anhxa import to_obj, to_json


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


MY_DIR = os.path.dirname(os.path.realpath(__file__))
KANGXI_FILE = os.path.join(MY_DIR, 'data', 'sino', 'kangxi.csv')

# -------------------------------------------------------------------------------
# Data Structures
# -------------------------------------------------------------------------------

KANGXI_FIELDS = ["idseq", "radical", "variants", "strokes", "meaning", "pinyin", "hanviet", "hiragana", "romaji", "hangeul", "romaja", "frequency", "simplified", "examples"]


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------

class Radical(object):

    ''' Chinese Radical
        Source: https://en.wikipedia.org/wiki/Kangxi_radical#Table_of_radicals
    '''
    def __init__(self, idseq='', radical='', variants='', strokes='', meaning='', pinyin='', hanviet='', hiragana='', romaji='', hangeul='', romaja='', frequency='', simplified='', examples=''):
        self.idseq = idseq
        self.radical = radical
        self.variants = variants
        self.strokes = strokes
        self.meaning = meaning
        self.pinyin = pinyin
        self.hanviet = hanviet
        self.hiragana = hiragana
        self.romaji = romaji
        self.hangeul = hangeul
        self.romaja = romaja
        self.frequency = frequency
        self.simplified = simplified
        self.examples = examples

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "{}-{}[sc:{}]".format(self.radical, self.meaning, self.strokes)

    def to_json(self):
        return to_json(self)

    __KANGXI_MAP = None

    @staticmethod
    def kangxi():
        if not Radical.__KANGXI_MAP:
            kxs = chio.read_csv(KANGXI_FILE, fieldnames=True)
            Radical.__KANGXI_MAP = KangxiMap(kxs)
        else:
            getLogger().debug("Kangxi has been loaded once. Created KangxiMap will be re-used")
        return Radical.__KANGXI_MAP


class KangxiMap:

    def __init__(self, rads=None):
        self.radicals = []
        self.rad_map = {}     # kangxi.radical -> kangxi object
        self.id_rad_map = {}  # idseq ('1', '2', i.e. string) -> rad object
        self.strokes_map = dd(list)  # map strokes => radicals
        if rads:
            for rad in rads:
                rad_obj = to_obj(Radical, rad)
                rad_obj.frequency = int(rad_obj.frequency)
                rad_obj.idseq = int(rad_obj.idseq)
                rad_obj.strokes = int(rad_obj.strokes)
                self.add(rad_obj)

    @property
    def all(self):
        return [r.radical for r in self.radicals]

    @property
    def strokes(self):
        return {sc: [r.radical for r in rads] for sc, rads in self.strokes_map.items()}

    def __len__(self):
        return len(self.radicals)

    def __getitem__(self, key):
        if key in self.rad_map:  # literal matching
            return self.rad_map[key]
        elif key in self.id_rad_map:
            return self.id_rad_map[key]
        else:
            return self.radicals[key]  # by list index

    def __contains__(self, key):
        return key in self.rad_map or key in self.id_rad_map

    def add(self, rad):
        self.radicals.append(rad)
        self.rad_map[rad.radical] = rad
        self.id_rad_map[str(rad.idseq)] = rad
        self.strokes_map[int(rad.strokes)].append(rad)
        # map variants & simplified
        if rad.variants:
            for v in rad.variants.split():
                self.rad_map[v] = rad
        if rad.simplified:
            for s in rad.simplified.split():
                self.rad_map[s] = rad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Script for testing leutile

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import logging
import unittest
from pathlib import Path

from chirptext.leutile import Counter, TextReport, StringTool, LOREM_IPSUM, Timer, piter
from chirptext.leutile import FileHelper
from chirptext.leutile import AppConfig


# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

TEST_DATA = os.path.join(os.path.dirname(__file__), 'data')


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Tests
# ------------------------------------------------------------------------------

class TestLeUtile(unittest.TestCase):

    def test_string_tool(self):
        self.assertEqual(StringTool.strip(None), '')
        self.assertEqual(StringTool.strip(' '), '')
        self.assertEqual(StringTool.to_str(None), '')
        # detokenize
        words = ["I", "'ll", "go", "home", "."]
        self.assertEqual(StringTool.detokenize(words), "I'll go home.")
        self.assertEqual(StringTool.detokenize(["This", "(", "thing", ")", "is", "a", "comment", "!"]), "This (thing) is a comment!")
        self.assertEqual(StringTool.detokenize("He said `` why ? '' .".split()), "He said “why?”.")
        self.assertEqual(StringTool.detokenize("Where are you ?".split()), "Where are you?")
        self.assertEqual(StringTool.detokenize("Note : It works .".split()), "Note: It works.")
        self.assertEqual(StringTool.detokenize("( A ) ; ".split()), "(A);")
        self.assertEqual(StringTool.detokenize("( A ) ; B ".split()), "(A); B")

    def test_piter(self):
        p1 = piter(range(5))
        l1 = [(x, p1.peep().value if p1.peep() else None) for x in p1]
        self.assertEqual(l1, [(0, 1), (1, 2), (2, 3), (3, 4), (4, None)])
        # empty list
        p2 = piter([])
        self.assertRaises(StopIteration, lambda: next(p2))
        self.assertIsNone(p2.current())
        self.assertIsNone(p2.peep())
        # 1 element iteration
        p3 = piter([1])
        v = next(p3)
        self.assertEqual(v, 1)
        self.assertIsNone(p3.peep())
        self.assertRaises(Exception, lambda: piter(None))
        # simple parser
        lssv = "src: me\ntxt:It works\n\nsrc:him\ntxt:It rains\ncmt:I made it up\n\nsrc:anonymous\ntxt:cow level is real"
        groups = []
        current = []
        p = piter(lssv.splitlines())
        for l in p:
            if l:
                current.append(l)
            if not l or not p.peep():
                if current:
                    groups.append(current)
                    current = []
        self.assertEqual(groups, [['src: me', 'txt:It works'], ['src:him', 'txt:It rains', 'cmt:I made it up'], ['src:anonymous', 'txt:cow level is real']])

    def test_counter(self):
        print("Test counter")
        c = Counter()
        c.count("A")
        c.count(None)
        c.count(None)
        self.assertEqual(c['A'], 1)
        self.assertEqual(c[None], 2)
        self.assertEqual(c.most_common(), [(None, 2), ('A', 1)])
        ct = Counter()
        for char in LOREM_IPSUM:
            if char == ' ':
                continue
            ct.count(char)
        top5chars = {(char, count) for char, count in ct.most_common(5)}
        expected = {('o', 29), ('e', 37), ('a', 29), ('t', 32), ('i', 42)}
        for k, v in ct.most_common(5):
            getLogger().debug("{}: {}".format(k, v))
        self.assertEqual(top5chars, expected)

    def test_textreport(self):
        with TextReport.null() as rp:
            rp.writeline("null")
            rp.writeline(123)
            self.assertEqual(rp.content(), '')
        with TextReport() as rp:
            rp.writeline("Test writing to stdout ...")
            rp.writeline("multiple line output", 123)
            self.assertEqual(rp.content(), '')
        with TextReport(os.path.join(TEST_DATA, "del.me")) as rp:
            rp.writeline("ABC")
            rp.writeline(123)
            self.assertEqual(rp.content(), '')
        self.assertTrue(rp.closed)
        # test string report
        with TextReport.string() as rp:
            rp.writeline("ABC")
            rp.writeline(123, 456, 789)
            self.assertEqual(rp.content(), 'ABC\n123 456 789\n')

    def test_timer(self):
        rp = TextReport.string()
        t = Timer(report=rp)
        msg = "Do something expensive"
        t.start(msg)
        do_expensive()
        t.stop(msg)
        getLogger().debug(rp.content())
        self.assertIn("Started", rp.content())
        self.assertIn("Stopped", rp.content())
        # test do()
        rp = TextReport.string()
        t = Timer(report=rp)
        t.do(lambda: do_expensive(), desc=msg)
        self.assertIn("Started", rp.content())
        self.assertIn("Stopped", rp.content())
        getLogger().debug(rp.content())


def do_expensive(n=10000):
    s = TextReport.string()
    for i in range(n):
        s.print("This is string number #{}".format(i))
    return str(s.content())


class TestFileHelper(unittest.TestCase):

    def test_replace_ext(self):
        self.assertEqual(Path(FileHelper.replace_ext('../data/foo.xml', 'json')).as_posix(),
                         '../data/foo.json')
        self.assertEqual(Path(FileHelper.replace_ext('../data/foo', 'json')).as_posix(),
                         '../data/foo.json')
        self.assertEqual(Path(FileHelper.replace_ext('../data/foo.xml', '')).as_posix(),
                         '../data/foo')
        self.assertEqual(Path(FileHelper.replace_ext('../data/foo.xml', None)).as_posix(),
                         '../data/foo')
        self.assertRaises(Exception, lambda: FileHelper.replace_ext(None, None))
        self.assertRaises(Exception, lambda: FileHelper.replace_ext('', None))
        self.assertRaises(Exception, lambda: FileHelper.replace_ext((1, 2, 3, 4), None))

    def test_rename(self):
        self.assertEqual(Path(FileHelper.replace_name('/data/foo.xml', 'bar')),
                         Path('/data/bar.xml'))
        self.assertEqual(Path(FileHelper.replace_name('../data/foo', 'bar')),
                         Path('../data/bar'))
        self.assertEqual(Path(FileHelper.replace_name('../foo.xml', 'bar')),
                         Path('../bar.xml'))
        self.assertRaises(Exception, lambda: FileHelper.replace_name(None, None))
        self.assertRaises(Exception, lambda: FileHelper.replace_name('', None))
        self.assertRaises(Exception, lambda: FileHelper.replace_name((1, 2, 3, 4), None))
        self.assertRaises(Exception, lambda: FileHelper.replace_name("/usr/foo", None))
        self.assertRaises(Exception, lambda: FileHelper.replace_name("/usr/foo", ''))
        self.assertRaises(Exception, lambda: FileHelper.replace_name("/usr/foo", (1, 2, 3)))


class TestConfigFile(unittest.TestCase):

    def test_locate_config_file(self):
        cfg = AppConfig(name='foo', mode=AppConfig.JSON)
        actual = cfg.potentials()
        expected = ['./.foo', './.foo.json', './foo', './foo.json', './data/foo', './data/foo.json', './data/.foo', './data/.foo.json', '~/.foo', '~/.foo.json', '~/.foo/config', '~/.foo.json/config', '~/.foo/config.json', '~/.foo.json/config.json', '~/.config/foo', '~/.config/foo.json', '~/.config/.foo', '~/.config/.foo.json', '~/.config/foo/config', '~/.config/foo.json/config', '~/.config/foo/config.json', '~/.config/foo.json/config.json', '~/.config/foo/foo', '~/.config/foo.json/foo.json']
        self.assertEqual(actual, expected)
        # default mode is INI
        cfg_ini = AppConfig('chirptest', working_dir=os.path.dirname(__file__))
        self.assertEqual(cfg_ini.config.sections(), ['AUTHOR'])
        self.assertEqual(cfg_ini.config['DEFAULT']['package'], 'chirptext.test')
        self.assertEqual(cfg_ini.config['DEFAULT']['tester'], 'unittest')
        self.assertEqual(cfg_ini.config['AUTHOR']['name'], 'Le Tuan Anh')
        self.assertEqual(cfg_ini.config['AUTHOR']['tester'], 'unittest')
        self.assertEqual(cfg_ini.config.get('AUTHOR', 'desc', fallback='nothing'), 'nothing')
        # test writing config
        with TextReport.string() as strfile:
            cfg_ini.config['AUTHOR']['desc'] = 'An author'
            cfg_ini.config.write(strfile.file)
            self.assertIn('desc = An author', strfile.content())
        # use JSON


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

''' Text Annotation module

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import logging
import json
import csv
import warnings
from collections import namedtuple
from collections import defaultdict as dd
from collections import OrderedDict

from .anhxa import IDGenerator
from .leutile import FileHelper
from .anhxa import DataObject
from . import chio
from .chio import iter_tsv_stream


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

STD_DIALECT = 'excel-tab'
STD_QUOTING = csv.QUOTE_MINIMAL
MODE_TSV = 'tsv'
MODE_JSON = 'json'
TokenInfo = namedtuple("TokenInfo", ['text', 'sk'])


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Classes
# ------------------------------------------------------------------------------

class Tag(DataObject):

    ''' A general tag which can be used for annotating different linguistic feature '''
    NONE = ''
    DEFAULT = 'n/a'
    GOLD = 'gold'
    ISF = 'isf'
    LELESK = 'lelesk'
    NTUMC = 'ntumc'
    MFS = 'mfs'
    MECAB = 'mecab'
    NLTK = 'nltk'
    WORDNET = 'wn'
    BABELNET = 'bn'
    OTHER = 'other'

    def __init__(self, label='', cfrom=-1, cto=-1, tagtype='', source=NONE, **kwargs):
        super().__init__(**kwargs)
        self.__cfrom = cfrom
        self.__cto = cto
        self.label = label
        self.source = source
        self.tagtype = tagtype  # tag type

    @property
    def cfrom(self):
        return self.__cfrom

    @cfrom.setter
    def cfrom(self, value):
        self.__cfrom = int(value) if value is not None else None

    @property
    def cto(self):
        return self.__cto

    @cto.setter
    def cto(self, value):
        self.__cto = int(value) if value is not None else None

    @property
    def type(self):
        ''' Alias for tagtype '''
        return self.tagtype

    @type.setter
    def type(self, value):
        self.tagtype = value

    @property
    def text(self):
        ''' Alias for label '''
        return self.label

    @text.setter
    def text(self, value):
        self.label = value

    def __repr__(self):
        if not self.tagtype:
            return '`{}`'.format(self.label)
        else:
            return '`{}:{}`'.format(self.tagtype, self.label)

    def __str__(self):
        return "`{l}`<{f}:{t}>{v}".format(l=self.label, f=self.cfrom, t=self.cto, v=self.tagtype)

    def to_json(self, default_cfrom=-1, default_cto=-1):
        a_dict = {'label': self.label}
        if self.tagtype:
            a_dict['type'] = self.tagtype
        if self.source:
            a_dict['source'] = self.source
        if self.cfrom is not None and self.cfrom != default_cfrom and self.cfrom >= 0:
            a_dict['cfrom'] = self.cfrom
        if self.cto is not None and self.cto != default_cto and self.cto >= 0:
            a_dict['cto'] = self.cto
        return a_dict

    @staticmethod
    def from_json(json_dict):
        tag = Tag()
        tag.update(json_dict, 'cfrom', 'cto', 'label', 'source', 'type')
        return tag


class Sentence(DataObject):

    def __init__(self, text='', ID=None, tokens=None, **kwargs):
        super().__init__(**kwargs)
        self.text = text
        self.ID = ID
        self.__tags = []
        self.__tokens = []
        self.__concepts = []
        self.__concept_map = OrderedDict()  # concept.ID to concept object
        if tokens:
            self.tokens = tokens

    @property
    def ID(self):
        return self.__ID

    @ID.setter
    def ID(self, value):
        self.__ID = int(value) if value else None

    def __repr__(self):
        return str(self)

    def __str__(self):
        # return format_tag(self)
        if self.ID:
            return "#{id}: {txt}".format(id=self.ID, txt=self.text)
        else:
            return self.text

    def __getitem__(self, idx):
        return self.__tokens[int(idx)]

    def __len__(self):
        return len(self.__tokens)

    @property
    def tags(self):
        ''' Sentence level tags '''
        return self.__tags

    def tagmap(self):
        tm = dd(list)
        for t in self.tags:
            tm[t.tagtype].append(t)
        return tm

    @property
    def tokens(self):
        return self.__tokens

    @tokens.setter
    def tokens(self, tokens):
        if self.__tokens:
            raise Exception("Cannot import tokens as my token list is not empty")
        else:
            self.import_tokens(tokens)

    @property
    def concepts(self):
        return self.__concepts

    def tcmap(self):
        ''' Create a tokens-concepts map '''
        tcmap = dd(list)
        for concept in self.__concept_map.values():
            for w in concept.tokens:
                tcmap[w].append(concept)
        return tcmap

    def mwe(self):
        ''' Return a generator of concepts that are linked to more than 1 token. '''
        return (c for c in self.__concepts if len(c.tokens) > 1)

    def msw(self):
        ''' Return a generator of tokens with more than one sense. '''
        return (t for t, c in self.tcmap().items() if len(c) > 1)

    def surface(self, tag):
        ''' Get surface string that is associated with a tag object '''
        if tag.cfrom is not None and tag.cto is not None and tag.cfrom >= 0 and tag.cto >= 0:
            return self.text[tag.cfrom:tag.cto]
        else:
            return ''

    def new_tag(self, label, cfrom=-1, cto=-1, tagtype='', **kwargs):
        ''' Create a sentence-level tag '''
        tag_obj = Tag(label, cfrom, cto, tagtype=tagtype, **kwargs)
        return self.add_tag(tag_obj)

    def get_tag(self, tagtype, auto_create=False, **kwargs):
        ''' Get the first tag with a type in this sentence
            use get_tag('mytype', default='somevalue') to get a new tag with default value 
            when there is no tag with this type. This new tag object will NOT be stored in the current sentence by default.
           
            use auto_create=True to auto create a new tag in the current sentence using the 'default' value.
            If there is no default value provided, an empty string '' will be used.
        '''
        for t in self.__tags:
            if t.tagtype == tagtype:
                return t
        if auto_create:
            return self.new_tag(label='' if 'default' not in kwargs else kwargs['default'], tagtype=tagtype, **kwargs)
        elif 'default' in kwargs:
            return Tag(label=kwargs['default'], tagtype=tagtype, **kwargs)
        else:
            raise LookupError("Sentence {} was not tagged with the speficied tagtype ({})".format(self, tagtype))

    def get_tags(self, tagtype, **kwargs):
        ''' Get all tags of a type '''
        return [t for t in self.__tags if t.tagtype == tagtype]

    def add_tag(self, tag_obj):
        ''' Add an existing tag object into this sentence '''
        self.tags.append(tag_obj)
        return tag_obj

    def new_token(self, *args, **kwargs):
        tk = Token(*args, **kwargs)
        return self.add_token_object(tk)

    def add_token_object(self, token):
        ''' Add a token object into this sentence '''
        token.sent = self  # take ownership of given token
        self.__tokens.append(token)
        return token

    def new_concept_id(self):
        ID = 0
        while ID in self.__concept_map:
            ID += 1
        return ID

    def new_concept(self, tag, clemma="", tokens=None, cidx=None, **kwargs):
        ''' Create a new concept object and add it to concept list
        tokens can be a list of Token objects or token indices
        '''
        if cidx is None:
            cidx = self.new_concept_id()
        if tokens:
            tokens = (t if isinstance(t, Token) else self[t] for t in tokens)
        c = Concept(cidx=cidx, tag=tag, clemma=clemma, sent=self, tokens=tokens, **kwargs)
        return self.add_concept(c)

    def add_concept(self, concept_obj):
        ''' Add a concept to current concept list '''
        if concept_obj is None:
            raise Exception("Concept object cannot be None")
        elif concept_obj in self.__concepts:
            raise Exception("Concept object is already inside")
        elif concept_obj.cidx in self.__concept_map:
            raise Exception("Duplicated concept ID ({})".format(concept_obj.cidx))
        self.__concepts.append(concept_obj)
        self.__concept_map[concept_obj.cidx] = concept_obj
        concept_obj.sent = self
        return concept_obj

    def pop_concept(self, cid, **kwargs):
        if cid not in self.__concept_map:
            if 'default' in kwargs:
                return kwargs['default']
            else:
                raise KeyError("Invalid cid")
        concept_obj = self.concept(cid)
        self.__concept_map.pop(cid)
        self.__concepts.remove(concept_obj)
        return concept_obj

    def concept(self, cid, **kwargs):
        ''' Get concept by concept ID '''
        if cid not in self.__concept_map:
            if 'default' in kwargs:
                return kwargs['default']
            else:
                raise KeyError("Invalid cid")
        else:
            return self.__concept_map[cid]

    def to_json(self):
        sent_dict = {'text': self.text}
        if self.tokens:
            sent_dict['tokens'] = [t.to_json() for t in self.tokens]
        if self.concepts:
            sent_dict['concepts'] = [c.to_json() for c in self.concepts]
        if self.ID is not None:
            sent_dict['ID'] = self.ID
        if self.flag is not None:
            sent_dict['flag'] = self.flag
        if self.comment is not None:
            sent_dict['comment'] = self.comment
        if self.__tags:
            sent_dict['tags'] = [t.to_json() for t in self.__tags]
        return sent_dict

    def import_tokens(self, tokens, import_hook=None, ignorecase=True):
        ''' Import a list of string as tokens '''
        text = self.text.lower() if ignorecase else self.text
        has_hooker = import_hook and callable(import_hook)
        cfrom = 0
        if self.__tokens:
            for tk in self.__tokens:
                if tk.cfrom and tk.cfrom > cfrom:
                    cfrom = tk.cfrom
        for token in tokens:
            if has_hooker:
                import_hook(token)
            to_find = token.lower() if ignorecase else token
            start = text.find(to_find, cfrom)
            # stanford parser
            if to_find == '``' or to_find == "''":
                start_dq = text.find('"', cfrom)
                if start_dq > -1 and (start == -1 or start > start_dq):
                    to_find = '"'
                    start = start_dq
            if to_find == '`' or to_find == "'":
                start_dq = text.find("'", cfrom)
                if start_dq > -1 and (start == -1 or start > start_dq):
                    to_find = "'"
                    start = start_dq
            if start == -1:
                raise LookupError('Cannot find token `{t}` in sent `{s}`({l}) from {i} ({p})'.format(t=token, s=self.text, l=len(self.text), i=cfrom, p=self.text[cfrom:cfrom + 20]))
            cfrom = start
            cto = cfrom + len(to_find)
            self.new_token(token, cfrom, cto)
            cfrom = cto - 1

    def fix_cfrom_cto(self, import_hook=None, ignorecase=True):
        text = self.text.lower() if ignorecase else self.text
        has_hooker = import_hook and callable(import_hook)
        cfrom = 0
        for token in self.tokens:
            if has_hooker:
                import_hook(token.text)
            start = text.find(token.text.lower() if ignorecase else token.text, cfrom)
            if start == -1:
                raise LookupError('Cannot find token `{t}` in sent `{s}`({l}) from {i} ({p})'.format(t=token, s=self.text, l=len(self.text), i=cfrom, p=self.text[cfrom:cfrom + 20]))
            cfrom = start
            cto = cfrom + len(token.text)
            token.cfrom = cfrom
            token.cto = cto
            cfrom = cto - 1

    @staticmethod
    def from_json(json_sent):
        sent = Sentence(json_sent['text'])
        sent.update(json_sent, 'ID', 'comment', 'flag')
        # import tokens
        for json_token in json_sent.get('tokens', []):
            sent.add_token_object(Token.from_json(json_token))
        # import concepts
        for json_concept in json_sent.get('concepts', []):
            tag = json_concept['tag']
            clemma = json_concept['clemma']
            tokenids = json_concept['tokens']
            concept = sent.new_concept(tag, clemma=clemma, tokens=tokenids)
            concept.update(json_concept, Concept.COMMENT, Concept.FLAG)
        for json_tag in json_sent.get('tags', []):
            sent.add_tag(Tag.from_json(json_tag))
        return sent


class Token(DataObject):

    def __init__(self, text='', cfrom=-1, cto=-1, sent=None, pos=None, lemma=None, comment=None, **kwargs):
        ''' A token (e.g. a word in a sentence) '''
        super().__init__(**kwargs)
        self.sent = sent
        self.__tags = []
        self.cfrom = cfrom
        self.cto = cto
        self.text = text  # original form
        self.lemma = lemma   # dictionary form
        self.pos = pos
        self.comment = comment

    def __getitem__(self, idx):
        return self.__tags[idx]

    def __len__(self):
        return len(self.__tags)

    def __iter__(self):
        return iter(self.__tags)

    @property
    def tags(self):
        return self.__tags

    def surface(self):
        if self.sent and self.sent.text:
            return self.sent.text[self.cfrom:self.cto]
        elif self.text:
            return self.text
        else:
            return ''

    def __repr__(self):
        return "`{l}`<{f}:{t}>".format(l=self.text, f=self.cfrom, t=self.cto)

    def tag_map(self):
        ''' Build a map from tagtype to list of tags '''
        tm = dd(list)
        for tag in self.__tags:
            tm[tag.tagtype].append(tag)
        return tm

    def __str__(self):
        return "`{l}`<{f}:{t}>{tgs}".format(l=self.text, f=self.cfrom, t=self.cto, tgs=self.__tags if self.__tags else '')

    def new_tag(self, label, cfrom=None, cto=None, tagtype=None, **kwargs):
        ''' Create a new tag on this token '''
        if cfrom is None:
            cfrom = self.cfrom
        if cto is None:
            cto = self.cto
        tag = Tag(label=label, cfrom=cfrom, cto=cto, tagtype=tagtype, **kwargs)
        return self.add_tag(tag)

    def get_tag(self, tagtype, auto_create=False, **kwargs):
        ''' Get the first tag with a type in this token
            use get_tag('mytype', default='somevalue') to get a new tag with default value 
            when there is no tag with this type. This new tag object will NOT be stored in the token by default.
           
            use auto_create=True to auto create a new tag in the current token using the 'default' value.
            If there is no default value provided, an empty string '' will be used.
        '''
        for t in self.__tags:
            if t.tagtype == tagtype:
                return t
        if auto_create:
            return self.new_tag(label='' if 'default' not in kwargs else kwargs['default'], tagtype=tagtype, **kwargs)
        elif 'default' in kwargs:
            return Tag(label=kwargs['default'], tagtype=tagtype, **kwargs)
        else:
            raise LookupError("Token {} is not tagged with the speficied tagtype ({})".format(self, tagtype))

    def get_tags(self, tagtype, **kwargs):
        ''' Get all token-level tags with the specified tagtype '''
        return [t for t in self.__tags if t.tagtype == tagtype]

    def add_tag(self, tag_obj):
        ''' Add an existing tag object into this token '''
        self.__tags.append(tag_obj)
        return tag_obj

    def find(self, tagtype, **kwargs):
        ''' (Deprecated) Return the first tag with a given tagtype in this token '''
        warnings.warn("Token.find() is deprecated and will be removed in near future. Use Token.get_tag() instead", DeprecationWarning, stacklevel=2)
        return self.get_tag(tagtype, **kwargs)

    def find_all(self, tagtype, **kwargs):
        ''' (Deprecated) Find all token-level tags with the specified tagtype '''
        warnings.warn("Token.find_all() is deprecated and will be removed in near future. Use Token.get_tags() instead", DeprecationWarning, stacklevel=2)
        return self.get_tags(tagtype, **kwargs)

    def to_json(self):
        token_json = {'cfrom': self.cfrom,
                      'cto': self.cto,
                      'text': self.text}
        if self.lemma:
            token_json['lemma'] = self.lemma
        if self.pos:
            token_json['pos'] = self.pos
        if self.comment:
            token_json['comment'] = self.comment
        if self.flag:
            token_json['flag'] = self.flag
        all_tags = [t.to_json(default_cfrom=self.cfrom, default_cto=self.cto) for t in self.tags]
        if all_tags:
            token_json['tags'] = all_tags
        return token_json

    @staticmethod
    def from_json(token_dict):
        tk = Token()
        tk.update(token_dict, 'cfrom', 'cto', 'text', 'lemma', 'pos', 'comment')
        # rebuild tags
        for tag_json in token_dict.get('tags', []):
            tk.add_tag(Tag.from_json(tag_json))
        return tk


class Concept(DataObject):

    FLAG = 'flag'
    COMMENT = 'comment'
    NOT_MATCHED = 'E'

    def __init__(self, cidx=None, tag='', clemma='', sent=None, tokens=None, comment=None, **kwargs):
        super().__init__(**kwargs)
        self.cidx = cidx
        self.sent = sent
        self.__tokens = []
        self.clemma = clemma
        self.tag = tag
        self.flag = None
        self.comment = comment if comment else ''
        if tokens:
            self.__tokens.extend(tokens)

    @property
    def tokens(self):
        return self.__tokens

    def add_token(self, *tokens):
        for token in tokens:
            self.__tokens.append(token)

    def __repr__(self):
        return '<{t}:"{l}">'.format(l=self.clemma, t=self.tag)

    def __str__(self):
        return '<{t}:"{l}">({ws})'.format(l=self.clemma, t=self.tag, ws=self.__tokens)

    def to_json(self):
        if self.sent:
            # get token idx from sent
            tokens = [self.sent.tokens.index(t) for t in self.__tokens]
        else:
            tokens = [t.text for t in self.__tokens]
        cdict = {
            'clemma': self.clemma,
            'tag': str(self.tag),
            'tokens': tokens
        }
        if self.comment:
            cdict[Concept.COMMENT] = self.comment
        if self.flag:
            cdict[Concept.FLAG] = self.flag
        return cdict


class Document(DataObject):

    def __init__(self, name='', path='.', **kwargs):
        super().__init__(**kwargs)
        self.__path = FileHelper.abspath(path)
        self.__name = name
        self.__sents = []
        self.__sent_map = {}
        self.__idgen = IDGenerator(id_hook=self.has_id)  # for creating a new sentence without ID

    def new_id(self):
        warnings.warn("new_id() is deprecated and will be removed in near future.", DeprecationWarning, stacklevel=2)
        return next(self.__idgen)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @property
    def path(self):
        return self.__path

    @path.setter
    def path(self, value):
        self.__path = value

    @property
    def sent_path(self):
        return os.path.join(self.path, '{}_sents.txt'.format(self.name))

    @property
    def token_path(self):
        return os.path.join(self.path, '{}_tokens.txt'.format(self.name))

    @property
    def concept_path(self):
        return os.path.join(self.path, '{}_concepts.txt'.format(self.name))

    @property
    def link_path(self):
        return os.path.join(self.path, '{}_links.txt'.format(self.name))

    @property
    def tag_path(self):
        return os.path.join(self.path, '{}_tags.txt'.format(self.name))

    def __len__(self):
        return len(self.__sents)

    def __getitem__(self, idx):
        return self.__sents[idx]

    def get(self, sent_id, **kwargs):
        ''' If sent_id exists, remove and return the associated sentence object else return default.
        If no default is provided, KeyError will be raised.'''
        if sent_id is not None and not isinstance(sent_id, int):
            sent_id = int(sent_id)
        if sent_id is None or not self.has_id(sent_id):
            if 'default' in kwargs:
                return kwargs['default']
            else:
                raise KeyError("Invalid sentence ID ({})".format(sent_id))
        return self.__sent_map[sent_id]

    def has_id(self, sent_id):
        return int(sent_id) in self.__sent_map

    def add_sent(self, sent_obj):
        ''' Add a ttl.Sentence object to this document '''
        if sent_obj is None:
            raise Exception("Sentence object cannot be None")
        elif sent_obj.ID is None:
            # if sentID is None, create a new ID
            sent_obj.ID = next(self.__idgen)
        elif self.has_id(sent_obj.ID):
            raise Exception("Sentence ID {} exists".format(sent_obj.ID))
        self.__sent_map[sent_obj.ID] = sent_obj
        self.__sents.append(sent_obj)
        return sent_obj

    def new_sent(self, text, ID=None, **kwargs):
        ''' Create a new sentence and add it to this Document '''
        if ID is None:
            ID = next(self.__idgen)
        return self.add_sent(Sentence(text, ID=ID, **kwargs))

    def pop(self, sent_id, **kwargs):
        ''' If sent_id exists, remove and return the associated sentence object else return default.
        If no default is provided, KeyError will be raised.'''
        if sent_id is not None and not isinstance(sent_id, int):
            sent_id = int(sent_id)
        if not self.has_id(sent_id):
            if 'default' in kwargs:
                return kwargs['default']
            else:
                raise KeyError("Sentence ID {} does not exist".format(sent_id))
        else:
            # sentence exists ...
            sent_obj = self.get(sent_id)
            self.__sent_map.pop(sent_id)
            self.__sents.remove(sent_obj)
            return sent_obj

    def read(self):
        ''' Read tagged doc from mutliple files (sents, tokens, concepts, links, tags) '''
        warnings.warn("Document.read() is deprecated and will be removed in near future.", DeprecationWarning, stacklevel=2)
        with TxtReader.from_doc(self) as reader:
            reader.read(self)
        return self

    @staticmethod
    def read_ttl(path):
        ''' Helper function to read Document in TTL-TXT format (i.e. ${docname}_*.txt)
        E.g. Document.read_ttl('~/data/myfile') is the same as Document('myfile', '~/data/').read()
        '''
        warnings.warn("Document.read_ttl() is deprecated and will be removed in near future. Use read() instead", DeprecationWarning, stacklevel=2)
        doc_path = os.path.dirname(path)
        doc_name = os.path.basename(path)
        return Document(doc_name, doc_path).read()

    def write_ttl(self):
        ''' Helper function to write doc to TTL-TXT format '''
        with TxtWriter.from_doc(self) as writer:
            writer.write_doc(self)

    @staticmethod
    def from_json_file(path):
        warnings.warn("Document.from_json_file() is deprecated and will be removed in near future.", DeprecationWarning, stacklevel=2)
        return read_json(path)


class TxtReader(object):
    def __init__(self, sent_stream, token_stream, concept_stream, link_stream, tag_stream, doc_name='', doc_path='.'):
        self.sent_stream = sent_stream
        self.token_stream = token_stream
        self.concept_stream = concept_stream
        self.link_stream = link_stream
        self.tag_stream = tag_stream
        self.doc_name = doc_name
        self.doc_path = doc_path

    def sent_reader(self):
        return iter_tsv_stream(self.sent_stream) if self.sent_stream else None

    def token_reader(self):
        return iter_tsv_stream(self.token_stream) if self.token_stream else None

    def concept_reader(self):
        return iter_tsv_stream(self.concept_stream) if self.concept_stream else None

    def link_reader(self):
        return iter_tsv_stream(self.link_stream) if self.link_stream else None

    def tag_reader(self):
        return iter_tsv_stream(self.tag_stream) if self.tag_stream else None

    @staticmethod
    def from_path(path):
        doc_path = os.path.dirname(path)
        doc_name = os.path.basename(path)
        doc = Document(name=doc_name, path=doc_path)
        return TxtReader.from_doc(doc)

    @staticmethod
    def from_doc(doc, encoding='utf-8'):
        reader = TxtReader(sent_stream=open(doc.sent_path, mode='rt', encoding=encoding),
                           token_stream=open(doc.token_path, mode='rt', encoding=encoding) if doc.token_path else None,
                           concept_stream=open(doc.concept_path, mode='rt', encoding=encoding) if doc.concept_path else None,
                           link_stream=open(doc.link_path, mode='rt', encoding=encoding) if doc.link_path else None,
                           tag_stream=open(doc.tag_path, mode='rt', encoding=encoding) if doc.tag_path else None,
                           doc_name=doc.name,
                           doc_path=doc.path)
        return reader

    def close(self):
        self.sent_stream.close()
        self.token_stream.close()
        self.concept_stream.close()
        self.link_stream.close()
        self.tag_stream.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def read(self, doc=None):
        ''' Read tagged doc from mutliple files (sents, tokens, concepts, links, tags) '''
        if not self.sent_stream:
            raise Exception("There is no sentence data stream available")
        if doc is None:
            doc = Document(name=self.doc_name, path=self.doc_path)
        for row in self.sent_reader():
            if len(row) == 2:
                sid, text = row
                doc.new_sent(text.strip(), ID=sid)
            elif len(row) == 4:
                sid, text, flag, comment = row
                sent = doc.new_sent(text.strip(), ID=sid)
                sent.flag = flag
                sent.comment = comment
        # Read tokens if available
        if self.token_stream:
            # read all tokens first
            sent_tokens_map = dd(list)
            for token_row in self.token_reader():
                if len(token_row) == 6:
                    sid, wid, token, lemma, pos, comment = token_row
                else:
                    sid, wid, token, lemma, pos = token_row
                    comment = ''
                sid = int(sid)
                sent_tokens_map[sid].append((token, lemma, pos.strip(), wid, comment))
                # TODO: verify wid?
            # import tokens
            for sent in doc:
                sent_tokens = sent_tokens_map[sent.ID]
                sent.import_tokens([x[0] for x in sent_tokens])
                for ((tk, lemma, pos, wid, comment), token) in zip(sent_tokens, sent.tokens):
                    token.pos = pos
                    token.lemma = lemma
                    token.comment = comment
            # only read concepts if tokens are available
            if self.concept_stream:
                # read concepts
                for concept_row in self.concept_reader():
                    if len(concept_row) == 5:
                        sid, cid, clemma, tag, comment = concept_row
                    else:
                        sid, cid, clemma, tag = concept_row
                        comment = ''
                    cid = int(cid)
                    doc.get(sid).new_concept(tag.strip(), clemma=clemma, cidx=cid, comment=comment)
                # only read concept-token links if tokens and concepts are available
                for sid, cid, wid in self.link_reader():
                    sent = doc.get(sid)
                    cid = int(cid)
                    wid = int(wid.strip())
                    sent.concept(cid).add_token(sent[wid])
        # read sentence level tags
        if self.tag_stream:
            for row in self.tag_reader():
                if len(row) == 5:
                    sid, cfrom, cto, label, tagtype = row
                    wid = None
                if len(row) == 6:
                    sid, cfrom, cto, label, tagtype, wid = row
                if cfrom:
                    cfrom = int(cfrom)
                if cto:
                    cto = int(cto)
                if wid is None or wid == '':
                    doc.get(sid).new_tag(label, cfrom, cto, tagtype=tagtype)
                else:
                    doc.get(sid)[int(wid)].new_tag(label, cfrom, cto, tagtype=tagtype)
        return doc


class TxtWriter(object):
    def __init__(self, sent_stream, token_stream, concept_stream, link_stream, tag_stream, id_seed=1):
        self.sent_stream = sent_stream
        self.token_stream = token_stream
        self.concept_stream = concept_stream
        self.link_stream = link_stream
        self.tag_stream = tag_stream
        self.sent_writer = csv.writer(sent_stream, dialect=STD_DIALECT, quoting=STD_QUOTING)
        self.token_writer = csv.writer(token_stream, dialect=STD_DIALECT, quoting=STD_QUOTING)
        self.concept_writer = csv.writer(concept_stream, dialect=STD_DIALECT, quoting=STD_QUOTING)
        self.link_writer = csv.writer(link_stream, dialect=STD_DIALECT, quoting=STD_QUOTING)
        self.tag_writer = csv.writer(tag_stream, dialect=STD_DIALECT, quoting=STD_QUOTING)
        self.__idgen = IDGenerator(id_seed=id_seed)

    def write_sent(self, sent, **kwargs):
        flag = sent.flag if sent.flag is not None else ''
        comment = sent.comment if sent.comment is not None else ''
        sid = sent.ID if sent.ID is not None else next(self.__idgen)
        self.sent_writer.writerow((sid, sent.text, flag, comment))
        # write tokens
        for wid, token in enumerate(sent.tokens):
            self.token_writer.writerow((sid, wid, token.text or token.surface(), token.lemma, token.pos, token.comment))
        # write concepts & wclinks
        for cid, concept in enumerate(sent.concepts):
            # write concept
            self.concept_writer.writerow((sid, cid, concept.clemma, concept.tag, concept.comment))
            # write cwlinks
            for token in concept.tokens:
                wid = sent.tokens.index(token)
                self.link_writer.writerow((sid, cid, wid))
        # write tags
        for tag in sent.tags:
            self.tag_writer.writerow((sid, tag.cfrom, tag.cto, tag.label, tag.tagtype, ''))
        # write token-level tags
        for wid, token in enumerate(sent.tokens):
            for tag in token:
                self.tag_writer.writerow((sid, tag.cfrom, tag.cto, tag.label, tag.tagtype, wid))

    def write_doc(self, doc, **kwargs):
        for sent in doc:
            self.write_sent(sent)

    def close(self):
        self.sent_stream.close()
        self.token_stream.close()
        self.concept_stream.close()
        self.link_stream.close()
        self.tag_stream.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def from_doc(doc, encoding='utf-8', **kwargs):
        return TxtWriter(sent_stream=open(doc.sent_path, mode='wt', encoding=encoding),
                         token_stream=open(doc.token_path, mode='wt', encoding=encoding),
                         concept_stream=open(doc.concept_path, mode='wt', encoding=encoding),
                         link_stream=open(doc.link_path, mode='wt', encoding=encoding),
                         tag_stream=open(doc.tag_path, mode='wt', encoding=encoding), **kwargs)

    @staticmethod
    def from_path(path, **kwargs):
        doc_path = os.path.dirname(path)
        doc_name = os.path.basename(path)
        doc = Document(name=doc_name, path=doc_path)
        return TxtWriter.from_doc(doc, **kwargs)


class JSONWriter(object):
    def __init__(self, output_stream, id_seed=1, **kwargs):
        self.__output_stream = output_stream
        self.__idgen = IDGenerator(id_seed=id_seed)

    def write_sent(self, sent, ensure_ascii=False, **kwargs):
        if sent.ID is None:
            sent.ID = next(self.__idgen)
        self.__output_stream.write(json.dumps(sent.to_json(), ensure_ascii=ensure_ascii))
        self.__output_stream.write('\n')

    def write_doc(self, doc, ensure_ascii=False, **kwargs):
        for sent in doc:
            self.write_sent(sent, ensure_ascii=ensure_ascii)

    def close(self):
        try:
            if self.__output_stream is not None:
                self.__output_stream.flush()
                self.__output_stream.close()
                self.__output_stream = None
        except:
            getLogger().exception("Could not close JSONWriter's output stream properly")

    @staticmethod
    def from_path(path, id_seed=1, **kwargs):
        return JSONWriter(output_stream=chio.open(path, mode='wt', **kwargs), id_seed=id_seed)

    @staticmethod
    def from_doc(doc, **kwargs):
        doc_path = os.path.join(doc.path, doc.name + '.ttl.json')
        return JSONWriter.from_path(doc_path, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


# ------------------------------------------------------------------------------
# Helper functions
# ------------------------------------------------------------------------------
def read_json_iter(path):
    if not os.path.isfile(path):
        raise Exception("Document file could not be found: {}".format(path))
    with chio.open(path) as infile:
        for line in infile:
            j = json.loads(line)
            sent = Sentence.from_json(j)
            yield sent
    return


def read_json(path):
    if not os.path.isfile(path):
        raise Exception("Document file could not be found: {}".format(path))
    doc_name = os.path.splitext(os.path.basename(path))[0]
    doc_path = os.path.dirname(path)
    doc = Document(doc_name, path=doc_path)
    for sent in read_json_iter(path):
        doc.add_sent(sent)
    return doc


def write_json(path, doc, ensure_ascii=False, **kwargs):
    with JSONWriter.from_path(path) as writer:
        writer.write_doc(doc, ensure_ascii=ensure_ascii, **kwargs)


def read(path, mode='tsv'):
    ''' Helper function to read Document in TTL-TXT format (i.e. ${docname}_*.txt)
    E.g. read('~/data/myfile') is the same as Document('myfile', '~/data/').read()
    '''
    if mode == 'tsv':
        doc_path = os.path.dirname(path)
        doc_name = os.path.basename(path)
        doc = Document(name=doc_name, path=doc_path)
        reader = TxtReader.from_doc(doc)
        doc = reader.read(doc=doc)
        reader.close()
        return doc
    elif mode == 'json':
        return read_json(path)
    else:
        raise Exception("Invalid mode - [{}] was provided".format(mode))


def write(path, doc, mode=MODE_TSV, **kwargs):
    ''' Helper function to write doc to TTL-TXT format '''
    if mode == MODE_TSV:
        with TxtWriter.from_path(path) as writer:
            writer.write_doc(doc)
    elif mode == MODE_JSON:
        write_json(path, doc, **kwargs)



# -*- coding: utf-8 -*-

''' Miscellaneous tools for text processing

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import io
import logging
import sys
import time
import errno
import json
import configparser
import warnings
from collections import Counter as PythonCounter
from collections import OrderedDict

from itertools import zip_longest

from .chio import read_file, write_file


# -------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


LOREM_IPSUM = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."


###############################################################################

def confirm(msg):
    return input(msg).upper() in ['Y', 'YES', 'OK']


def uniquify(a_list):
    return list(OrderedDict(zip(a_list, range(len(a_list)))).keys())


def header(*msg, level='h1', separator=" ", print_out=print):
    ''' Print header block in text mode
    '''
    out_string = separator.join(str(x) for x in msg)
    if level == 'h0':
        # box_len = 80 if len(msg) < 80 else len(msg)
        box_len = 80
        print_out('+' + '-' * (box_len + 2))
        print_out("| %s" % out_string)
        print_out('+' + '-' * (box_len + 2))
    elif level == 'h1':
        print_out("")
        print_out(out_string)
        print_out('-' * 60)
    elif level == 'h2':
        print_out('\t%s' % out_string)
        print_out('\t' + ('-' * 40))
    else:
        print_out('\t\t%s' % out_string)
        print_out('\t\t' + ('-' * 20))


def is_number(s):
    ''' Check if something is a number
    '''
    try:
        if str(float(s)) != 'nan':
            return True
    except Exception:
        pass
    return False


def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


###############################################################################

class Value(object):
    ''' Value holder '''
    def __init__(self, value=None):
        self.value = value


class piter(object):
    ''' Peep-able iterator '''
    def __init__(self, iterable):
        self.__iterable = iter(iterable)
        self.__current = None
        self.__peep = None
        self.__ended = False

    def current(self):
        return self.__current

    def peep(self):
        return self.__peep

    def fetch(self, value_obj=None):
        ''' Fetch the next two values '''
        val = None
        try:
            val = next(self.__iterable)
        except StopIteration:
            return None
        if value_obj is None:
            value_obj = Value(value=val)
        else:
            value_obj.value = val
        return value_obj

    def __iter__(self):
        return self

    def __next__(self):
        if self.__peep is not None:
            self.__current.value = self.__peep.value  # move peep to current
        else:
            self.__current = self.fetch(self.__current)
        # peep
        if self.__current is None:
            self.__peep = None  # iteration ended
            raise StopIteration
        else:
            self.__peep = self.fetch(self.__peep)
            return self.__current.value


class Counter(PythonCounter):
    ''' Powerful counter class
    '''
    def __init__(self, priority=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__priority = priority if priority else []

    def count(self, key):
        self.update({key})

    @property
    def priority(self):
        return list(self.__priority)

    @priority.setter
    def priority(self, priority):
        self.priority = list(priority)

    def get_report_order(self):
        ''' Keys are sorted based on report order (i.e. some keys to be shown first)
            Related: see sorted_by_count
        '''
        order_list = []
        for x in self.__priority:
            order_list.append([x, self[x]])
        for x in sorted(list(self.keys())):
            if x not in self.__priority:
                order_list.append([x, self[x]])
        return order_list

    def summarise(self, report=None, byfreq=True, limit=None):
        if not report:
            report = TextReport()
        items = self.most_common() if byfreq else self.get_report_order()
        if limit:
            items = items[:limit]
        for k, v in items:
            report.writeline("%s: %d" % (k, v))

    def group_by_count(self):
        d = OrderedDict()
        for item, count in self.most_common():
            if count not in d:
                d[count] = []
            d[count].append(item)
        return d.items()


class Timer:
    ''' Measure tasks' runtime
    '''
    def __init__(self, logger=None, report=None):
        self.start_time = time.time()
        self.end_time = time.time()
        self.__logger = logger
        self.__report = report
        self.end = self.stop  # just an alias

    @property
    def logger(self):
        return self.__logger if self.__logger is not None else getLogger()

    def exec_time(self):
        ''' Calculate run time '''
        return self.end_time - self.start_time

    def log(self, action, desc=None):
        msg = '{action} - [{desc}]'.format(action=action, desc=desc) if desc else action
        self.logger.info(msg)
        if self.__report:
            self.__report.writeline(msg)
        return self

    def start(self, desc=''):
        self.log("Started", desc=desc)
        self.start_time = time.time()
        return self

    def stop(self, desc=''):
        self.end_time = time.time()
        msg = "[{} | {}]".format(desc, str(self)) if desc else str(self)
        self.log("Stopped", desc=msg)
        return self

    def do(self, task, desc=''):
        self.start(desc)
        task()
        self.stop(desc)

    def __str__(self):
        return "Execution time: %.2f sec(s)" % (self.exec_time(),)


###############################################################################

class StringTool:
    ''' Common string function
    '''
    @staticmethod
    def strip(a_str):
        return a_str.strip() if a_str else ''

    @staticmethod
    def to_str(a_str):
        return str(a_str) if a_str else ''

    @staticmethod
    def detokenize(tokens):
        sentence_text = ' '.join(tokens)
        sentence_text = sentence_text.replace(" , , ", ", ")
        sentence_text = sentence_text.replace(' , ', ', ').replace('“ ', '“').replace(' ”', '”')
        sentence_text = sentence_text.replace(' ! ', '! ').replace(" 'll ", "'ll ").replace(" 've ", "'ve ").replace(" 're ", "'re ").replace(" 'd ", "'d ")
        sentence_text = sentence_text.replace(" 's ", "'s ")
        sentence_text = sentence_text.replace(" 'm ", "'m ")
        sentence_text = sentence_text.replace(" ' ", "' ")
        sentence_text = sentence_text.replace(" ; ", "; ")
        sentence_text = sentence_text.replace(" : ", ": ")
        sentence_text = sentence_text.replace("( ", "(")
        sentence_text = sentence_text.replace(" )", ")")
        sentence_text = sentence_text.replace(" ?", "?")
        sentence_text = sentence_text.replace(" n't ", "n't ")
        sentence_text = sentence_text.replace("  ", " ")
        sentence_text = sentence_text.replace('``', "“").replace("''", "”")
        sentence_text = sentence_text.replace("“ ", "“").replace(" ”", "”")
        if sentence_text[-2:] in (' .', ' :', ' ?', ' !', " ;"):
            sentence_text = sentence_text[:-2] + sentence_text[-1]
        sentence_text = sentence_text.strip()
        return sentence_text


###############################################################################

class TextReport:

    STDOUT = '*stdout*'
    STRINGIO = '*string*'

    ''' Helper for creating text report with indentation, tables and flexible output (to stdout or a file)
    '''
    def __init__(self, path=None, mode='w', name=None, auto_flush=True, encoding='utf8'):
        ''' Create a text report.

        Arguments:
            report_path -- Path to report file
            mode        -- a for append, w (default) for create from scratch (overwrite existing file)
        '''
        if not path or path == TextReport.STDOUT:
            self.__path = TextReport.STDOUT
            self.__report_file = sys.stdout
            self.name = 'stdout'
            self.mode = None
            self.auto_flush = False
        elif path == TextReport.STRINGIO:
            self.__path = TextReport.STRINGIO
            self.__report_file = io.StringIO()
            self.name = 'StringIO'
            self.mode = None
            self.auto_flush = False
        else:
            if path == '/dev/null':
                self.__path = '/dev/null'
                self.__report_file = open(os.devnull, mode=mode, encoding=encoding)
            else:
                self.__path = os.path.expanduser(path)
                self.__report_file = open(self.__path, mode, encoding=encoding)
            self.name = name if name else FileHelper.getfilename(self.__path)
            self.auto_flush = auto_flush
            self.mode = mode
        self.print = self.writeline  # just an alias

    @property
    def closed(self):
        return self.__report_file is None or self.__report_file.closed

    @property
    def stream(self):
        return self.__report_file
    
    @property
    def file(self):
        ''' An alias for stream '''
        return self.__report_file

    def content(self):
        ''' Return report content as a string if mode == STRINGIO else an empty string '''
        if isinstance(self.__report_file, io.StringIO):
            return self.__report_file.getvalue()
        else:
            return ''

    def write(self, *msg, separator=" ", level=0):
        out_string = separator.join(str(x) for x in msg)
        self.__report_file.write("\t" * level)
        self.__report_file.write(out_string)
        if self.auto_flush:
            self.__report_file.flush()

    def writeline(self, *msg, **kwargs):
        self.write(*msg, **kwargs)
        self.write('\n')

    def header(self, *msg, **kwargs):
        header(*msg, print_out=self.writeline, **kwargs)

    def close(self):
        if self.mode and self.__report_file != sys.stdout:
            try:
                self.__report_file.flush()
                self.__report_file.close()
                self.__report_file = None
            except:
                getLogger().exception("Error raised while saving report")
                raise

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def path(self):
        return self.__path

    @staticmethod
    def null():
        ''' Get a dev null report (print to nowhere)'''
        return TextReport('/dev/null')

    @staticmethod
    def string(**kwargs):
        return TextReport(TextReport.STRINGIO, **kwargs)


###############################################################################

class FileHub:
    ''' A helper class for working with multiple text reports at the same time
    '''
    def __init__(self, *filenames, working_dir='.', default_mode='a', ext='txt'):
        self.files = {}
        self.ext = ext if ext else ''
        self.working_dir = working_dir
        self.default_mode = default_mode

    def __getitem__(self, key):
        if key not in self.files:
            self.open(key)
        return self.files[key]

    def __setitem__(self, key, value):
        self.files[key] = value

    def get_path(self, key):
        if os.path.isabs(key):
            return key
        else:
            return os.path.join(self.working_dir, key + '.' + self.ext)

    def open(self, key, mode=None, **kwargs):
        if not mode:
            mode = self.default_mode
        self.files[key] = TextReport(self.get_path(key), mode=mode, **kwargs)
        return self.files[key]

    def close(self):
        for key in self.files.keys():
            self[key].close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


###############################################################################

class Table:
    ''' A text-based table which can be used with TextReport
    '''
    def __init__(self, header=True, padding=True, NoneValue=None):
        self.rows = []
        self.col_count = 0
        self.NoneValue = NoneValue
        self.max_lengths = []
        self.header = header
        self.padding = padding

    def maxlength(self, val_a, val_b):
        len_a = len(val_a) if val_a else 0
        len_b = len(val_b) if val_b else 0
        return len_a if len_a > len_b else len_b

    def add_row(self, new_row):
        if new_row is None:
            raise ValueError("Row cannot be None")
        if len(new_row) > self.col_count:
            self.col_count = len(new_row)  # longer row
            for row in self.rows:
                # add cell
                row += [self.NoneValue] * (len(new_row) - self.col_count)
        elif len(new_row) < self.col_count:
            new_row += [self.NoneValue] * (self.col_count - len(new_row))
        self.rows.append(list(new_row))  # clone a list, rather than store the ref passed in

    def __getitem__(self, row_id):
        return list(self.rows[row_id])    # clone a row to return

    def get_column(self, col_id):
        return [x[col_id] if x[col_id] is not None else self.NoneValue for x in self.rows]

    def format(self):
        ''' Format table to print out
        '''
        self.max_lengths = []
        for row in self.rows:
            if len(self.max_lengths) < len(row):
                self.max_lengths += [0] * (len(row) - len(self.max_lengths))
            for idx, val in enumerate(row):
                len_cell = len(str(val)) if val else 0
                if self.max_lengths[idx] < len_cell:
                    self.max_lengths[idx] = len_cell
        return self.max_lengths

    def print_separator(self, print_func):
        self.print_cells(['-' * (x + (2 if self.padding else 0)) for x in self.max_lengths], print_func, joint='+')

    def print_cells(self, cells, print_func=None, extra_lines=False, joint='|'):
        if print_func is None:
            print(joint + joint.join(cells) + joint)
        else:
            print_func(joint + joint.join(cells) + joint)
        if extra_lines:
            self.print_separator(print_func)

    def print(self, print_func=print):
        max_lengths = self.format()
        self.print_separator(print_func)
        for ridx, row in enumerate(self.rows):
            cells = []
            for idx, cell in enumerate(row):
                txt = str(cell)
                cell_width = max_lengths[idx]
                if ridx == 0 and self.header:
                    txt = txt.center(cell_width)
                else:
                    txt = txt.rjust(cell_width) if is_number(txt) else txt.ljust(cell_width)
                if self.padding:
                    txt = ' ' + txt + ' '
                cells.append(txt)
            self.print_cells(cells, print_func, extra_lines=(self.header and ridx == 0))
        self.print_separator(print_func)


###############################################################################

class FileHelper:
    @staticmethod
    def getfilename(file_path):
        ''' Get filename without extension
        '''
        return os.path.splitext(os.path.basename(file_path))[0]

    @staticmethod
    def getfullfilename(file_path):
        ''' Get full filename (with extension)
        '''
        warnings.warn("getfullfilename() is deprecated and will be removed in near future. Use chirptext.io.write_file() instead", DeprecationWarning)
        if file_path:
            return os.path.basename(file_path)
        else:
            return ''

    @staticmethod
    def replace_ext(file_path, ext):
        ''' Change extension of a file_path to something else (provide None to remove) '''
        if not file_path:
            raise Exception("File path cannot be empty")
        dirname = os.path.dirname(file_path)
        filename = FileHelper.getfilename(file_path)
        if ext:
            filename = filename + '.' + ext
        return os.path.join(dirname, filename)

    @staticmethod
    def replace_name(file_path, new_name):
        ''' Change the file name in a path but keep the extension '''
        if not file_path:
            raise Exception("File path cannot be empty")
        elif not new_name:
            raise Exception("New name cannot be empty")
        dirname = os.path.dirname(file_path)
        ext = os.path.splitext(os.path.basename(file_path))[1]
        return os.path.join(dirname, new_name + ext)

    @staticmethod
    def abspath(a_path):
        return os.path.abspath(os.path.expanduser(a_path))

    @staticmethod
    def create_dir(dir_path):
        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)
            except:
                getLogger().exception("Cannot create folder [%s]" % (dir_path,))
                raise

    @staticmethod
    def get_child_folders(path):
        ''' Get all child folders of a folder '''
        path = FileHelper.abspath(path)
        return [dirname for dirname in os.listdir(path) if os.path.isdir(os.path.join(path, dirname))]

    @staticmethod
    def get_child_files(path):
        ''' Get all child files of a folder '''
        path = FileHelper.abspath(path)
        return [filename for filename in os.listdir(path) if os.path.isfile(os.path.join(path, filename))]

    @staticmethod
    def remove_file(filepath):
        ''' Delete a file '''
        try:
            os.remove(os.path.abspath(os.path.expanduser(filepath)))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    @staticmethod
    def save(path, content, *args, **kwargs):
        warnings.warn("save() is deprecated and will be removed in near future. Use chirptext.io.write_file() instead", DeprecationWarning, stacklevel=2)
        return write_file(path, content, *args, **kwargs)

    @staticmethod
    def read(path, mode='r', encoding='utf-8', *args, **kwargs):
        warnings.warn("read() is deprecated and will be removed in near future. Use chirptext.io.read_file() instead", DeprecationWarning, stacklevel=2)
        return read_file(path, mode=mode, encoding=encoding, *args, **kwargs)


class AppConfig(object):

    ''' Application Configuration Helper
This class supports guessing configuration file location, and reads either INI (default) or JSON format.
    '''
    JSON = 'json'
    INI = 'ini'  # Python INI config file
    LOC_TEMPLATE = ['{wd}/.{n}', '{wd}/{n}', '{wd}/data/{n}', '{wd}/data/.{n}', '~/.{n}',
                    '~/.{n}/config', '~/.{n}/config.{mode}', '~/.config/{n}',
                    '~/.config/.{n}', '~/.config/{n}/config', '~/.config/{n}/config.{mode}',
                    '~/.config/{n}/{n}']

    def __init__(self, name, mode=INI, working_dir='.', extra_potentials=None):
        self.__name = name
        self.__mode = mode
        self.working_dir = working_dir
        self.__potential = []
        if extra_potentials:
            self.add_potential(*extra_potentials)
        self.add_potential(*AppConfig.LOC_TEMPLATE)
        self.__config = None
        self.__config_path = None

    def potentials(self):
        return self.__potential

    def _ptn2fn(self, pattern):
        ''' Pattern to filename '''
        return [pattern.format(wd=self.working_dir, n=self.__name, mode=self.__mode),
                pattern.format(wd=self.working_dir, n='{}.{}'.format(self.__name, self.__mode), mode=self.__mode)]

    def add_potential(self, *patterns):
        ''' Add a potential config file pattern '''
        for ptn in patterns:
            self.__potential.extend(self._ptn2fn(ptn))

    def locate_config(self):
        ''' Locate config file '''
        for f in self.__potential:
            f = FileHelper.abspath(f)
            if os.path.isfile(f):
                return f
        return None

    def read_config(self, key, strict=False, **kwargs):
        ''' Read a config by key. Default value can be passed by using the kwarg `default`
            For example: read_config(key, default='my value')
            Arguments:
                default -- Optional kwarg to set default value when key could not be found
                strict -- Set to True to throw an exception if config key was not set. Defaulted to False
        '''
        if self.config and key in self.config:
            return self.config[key]
        elif 'default' in kwargs:
            return kwargs['default']
        elif strict:
            if not self.config:
                raise FileNotFoundError("Configuration file could not be loaded")
            else:
                raise KeyError("Key `{}` could not be found in config file: {}".format(key, self.__config_path))
    
    @property
    def config(self):
        ''' Read config automatically if required '''
        if self.__config is None:
            config_path = self.locate_config()
            if config_path:
                self.__config = self.read_file(config_path)
                self.__config_path = config_path
        return self.__config

    def read_file(self, file_path):
        ''' Read a configuration file and return configuration data '''
        getLogger().info("Loading app config from {} file: {}".format(self.__mode, file_path))
        if self.__mode == AppConfig.JSON:
            return json.loads(read_file(file_path), object_pairs_hook=OrderedDict)
        elif self.__mode == AppConfig.INI:
            config = configparser.ConfigParser(allow_no_value=True)
            config.read(file_path)
            return config

    def load(self, file_path):
        ''' Load configuration from a specific file '''
        self.clear()
        self.__config = self.read_file(file_path)

    def clear(self):
        self.__config = None
        return self



# -*- coding: utf-8 -*-

'''
Readers for popular formats

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import csv
import gzip
import logging
import warnings

# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

QUOTE_MINIMAL = csv.QUOTE_MINIMAL
QUOTE_NONE = csv.QUOTE_NONE
QUOTE_ALL = csv.QUOTE_ALL
__python_open = open


def getLogger():
    return logging.getLogger()


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------

def to_string(content, encoding='utf-8'):
    if isinstance(content, bytes):
        return content.decode(encoding)
    elif isinstance(content, str):
        return content
    else:
        return str(content)


def is_file(path):
    ''' Check if path is a path to an existing file '''
    return path and os.path.isfile(to_string(path))


def open(path, encoding='utf-8', mode='rt', *args, **kwargs):
    if not mode:
        raise Exception("Invalid file access mode")
    elif mode.startswith('r') and not is_file(path):
        raise FileNotFoundError("File {} does not exist".format(path))
    if 't' not in mode and 'b' not in mode:
        mode += 't'  # default all open to text mode
    if not path:
        raise ValueError("Path cannot be empty")
    # ensure that path is a str so it'll work on python 3.5 and prior
    if not isinstance(path, str):
        path = str(path)
    # read or write
    if path.endswith('.gz'):
        if mode.endswith('b'):
            return gzip.open(path, mode=mode)
        else:
            return gzip.open(path, mode=mode, encoding=encoding)
    else:
        if mode.endswith('b'):
            return __python_open(path, mode=mode)
        else:
            return __python_open(path, mode=mode, encoding=encoding)


def process_file(path, processor, encoding='utf-8', mode='rt', *args, **kwargs):
    ''' Process a text file's content. If the file name ends with .gz, read it as gzip file '''
    if mode not in ('rU', 'rt', 'rb', 'r'):
        raise Exception("Invalid file reading mode")
    with open(path, encoding=encoding, mode=mode, *args, **kwargs) as infile:
        return processor(infile)


def read_file(path, encoding='utf-8', *args, **kwargs):
    ''' Read text file content. If the file name ends with .gz, read it as gzip file.
    If mode argument is provided as 'rb', content will be read as byte stream.
    By default, content is read as string.
    '''
    if 'mode' in kwargs and kwargs['mode'] == 'rb':
        return process_file(path, processor=lambda x: x.read(),
                            encoding=encoding, *args, **kwargs)
    else:
        return process_file(path, processor=lambda x: to_string(x.read(), encoding),
                            encoding=encoding, *args, **kwargs)


def write_file(path, content, mode=None, encoding='utf-8'):
    ''' Write content to a file. If the path ends with .gz, gzip will be used. '''
    if not mode:
        if isinstance(content, bytes):
            mode = 'wb'
        else:
            mode = 'wt'
    if not path:
        raise ValueError("Output path is invalid")
    else:
        getLogger().debug("Writing content to {}".format(path))
        # convert content to string when writing text data
        if mode in ('w', 'wt') and not isinstance(content, str):
            content = to_string(content)
        elif mode == 'wb':
            # content needs to be encoded as bytes
            if not isinstance(content, str):
                content = to_string(content).encode(encoding)
            else:
                content = content.encode(encoding)
        if str(path).endswith('.gz'):
            with gzip.open(path, mode) as outfile:
                outfile.write(content)
        else:
            with open(path, mode=mode) as outfile:
                outfile.write(content)


def iter_csv_stream(input_stream, fieldnames=None, sniff=False, *args, **kwargs):
    ''' Read CSV content as a table (list of lists) from an input stream '''
    if 'dialect' not in kwargs and sniff:
        kwargs['dialect'] = csv.Sniffer().sniff(input_stream.read(1024))
        input_stream.seek(0)
    if 'quoting' in kwargs and kwargs['quoting'] is None:
        kwargs['quoting'] = csv.QUOTE_MINIMAL
    if fieldnames:
        # read csv using dictreader
        if isinstance(fieldnames, bool):
            reader = csv.DictReader(input_stream, *args, **kwargs)
        else:
            reader = csv.DictReader(input_stream, *args, fieldnames=fieldnames, **kwargs)
        for row in reader:
            yield row
    else:
        csvreader = csv.reader(input_stream, *args, **kwargs)
        for row in csvreader:
            yield row


def iter_tsv_stream(input_stream, *args, **kwargs):
    return iter_csv_stream(input_stream, *args, dialect='excel-tab', **kwargs)


def read_csv_iter(path, fieldnames=None, sniff=True, mode='rt', encoding='utf-8', *args, **kwargs):
    ''' Iterate through CSV rows in a file.
    By default, csv.reader() will be used any output will be a list of lists.
    If fieldnames is provided, DictReader will be used and output will be list of OrderedDict instead.
    CSV sniffing (dialect detection) is enabled by default, set sniff=False to switch it off.
    '''
    with open(path, mode=mode, encoding=encoding) as infile:
        for row in iter_csv_stream(infile, fieldnames=fieldnames, sniff=sniff, *args, **kwargs):
            yield row


def read_tsv_iter(path, *args, **kwargs):
    return read_csv_iter(path, *args, dialect='excel-tab', **kwargs)


def read_csv(path, fieldnames=None, sniff=True, encoding='utf-8', *args, **kwargs):
    ''' Read CSV rows as table from a file.
    By default, csv.reader() will be used any output will be a list of lists.
    If fieldnames is provided, DictReader will be used and output will be list of OrderedDict instead.
    CSV sniffing (dialect detection) is enabled by default, set sniff=False to switch it off.
    '''
    return list(r for r in read_csv_iter(path, fieldnames=fieldnames, sniff=sniff, encoding=encoding, *args, **kwargs))


def read_tsv(path, *args, **kwargs):
    return read_csv(path, dialect='excel-tab', *args, **kwargs)


def write_csv(path, rows, dialect='excel', fieldnames=None, quoting=csv.QUOTE_ALL, extrasaction='ignore', *args, **kwargs):
    ''' Write rows data to a CSV file (with or without fieldnames) '''
    if not quoting:
        quoting = csv.QUOTE_MINIMAL
    if 'lineterminator' not in kwargs:
        kwargs['lineterminator'] = '\n'  # use \n to fix double-line in Windows
    with open(path, mode='wt', newline='') as csvfile:
        if fieldnames:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames, dialect=dialect, quoting=quoting, extrasaction=extrasaction, *args, **kwargs)
            writer.writeheader()
            for row in rows:
                writer.writerow(row)
        else:
            writer = csv.writer(csvfile, dialect=dialect, quoting=quoting, *args, **kwargs)
            for row in rows:
                writer.writerow(row)


def write_tsv(path, rows, *args, **kwargs):
    return write_csv(path, rows, dialect='excel-tab', *args, **kwargs)


class CSV(object):

    QUOTE_MINIMAL = csv.QUOTE_MINIMAL
    QUOTE_NONE = csv.QUOTE_NONE
    QUOTE_ALL = csv.QUOTE_ALL

    @staticmethod
    def read(file_name, header=False, *args, **kwargs):
        warnings.warn("chirptext.chio.CSV is deprecated and will be removed in near future. Use chio.read_csv() instead.", DeprecationWarning, stacklevel=2)
        return read_csv(file_name, fieldnames=header, *args, **kwargs)

    @staticmethod
    def read_tsv(file_name, *args, **kwargs):
        warnings.warn("chirptext.chio.CSV is deprecated and will be removed in near future. Use chio.read_tsv() instead.", DeprecationWarning, stacklevel=2)
        return read_tsv(file_name, *args, **kwargs)

    @staticmethod
    def write(file_name, rows, header=None, *args, **kwargs):
        ''' Write rows data to a CSV file (with or without header) '''
        warnings.warn("chirptext.io.CSV is deprecated and will be removed in near future. Use chio.write_csv() instead.", DeprecationWarning, stacklevel=2)
        write_csv(file_name, rows, fieldnames=header, *args, **kwargs)

    @staticmethod
    def write_tsv(file_name, rows, *args, **kwargs):
        warnings.warn("chirptext.io.CSV is deprecated and will be removed in near future. Use chio.write_tsv() instead.", DeprecationWarning, stacklevel=2)
        write_tsv(file_name, rows, *args, **kwargs)



# -*- coding: utf-8 -*-

'''
Leipzig standard abbreviations

Latest version can be found at https://github.com/letuananh/chirptext

References:
    The Leipzig Glossing Rules
        https://www.eva.mpg.de/lingua/pdf/Glossing-Rules.pdf

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import logging

from .anhxa import DataObject


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Data Structures
# -------------------------------------------------------------------------------

class SmartTag(DataObject):

    def __init__(self, label, description='', **kwargs):
        super().__init__(**kwargs)
        self.label = label
        self.description = description

    def __repr__(self):
        return "SmartTag(label={})".format(repr(self.label))

    def __str__(self):
        return str(self.label) if self.label is not None else ''


# Adopted from version: Leipzig, last change: May 31, 2015
# Name changed:
#    + 1, 2, 3 -> _1, _2, _3
#    + N- -> NON
_1 = SmartTag('1', 'first person')
_2 = SmartTag('2', 'second person')
_3 = SmartTag('3', 'third person')
A = SmartTag('A', 'agent-like argument of canonical transitive verb')
ABL = SmartTag('ABL', 'ablative')
ABS = SmartTag('ABS', 'absolutive')
ACC = SmartTag('ACC', 'accusative')
ADJ = SmartTag('ADJ', 'adjective')
ADV = SmartTag('ADV', 'adverb(ial)')
AGR = SmartTag('AGR', 'agreement')
ALL = SmartTag('ALL', 'allative')
ANTIP = SmartTag('ANTIP', 'antipassive')
APPL = SmartTag('APPL', 'applicative')
ART = SmartTag('ART', 'article')
AUX = SmartTag('AUX', 'auxiliary')
BEN = SmartTag('BEN', 'benefactive')
CAUS = SmartTag('CAUS', 'causative')
CLF = SmartTag('CLF', 'classifier')
COM = SmartTag('COM', 'comitative')
COMP = SmartTag('COMP', 'complementizer')
COMPL = SmartTag('COMPL', 'completive')
COND = SmartTag('COND', 'conditional')
COP = SmartTag('COP', 'copula')
CVB = SmartTag('CVB', 'converb')
DAT = SmartTag('DAT', 'dative')
DECL = SmartTag('DECL', 'declarative')
DEF = SmartTag('DEF', 'definite')
DEM = SmartTag('DEM', 'demonstrative')
DET = SmartTag('DET', 'determiner')
DIST = SmartTag('DIST', 'distal')
DISTR = SmartTag('DISTR', 'distributive')
DU = SmartTag('DU', 'dual')
DUR = SmartTag('DUR', 'durative')
ERG = SmartTag('ERG', 'ergative')
EXCL = SmartTag('EXCL', 'exclusive')
F = SmartTag('F', 'feminine')
FOC = SmartTag('FOC', 'focus')
FUT = SmartTag('FUT', 'future')
GEN = SmartTag('GEN', 'genitive')
IMP = SmartTag('IMP', 'imperative')
INCL = SmartTag('INCL', 'inclusive')
IND = SmartTag('IND', 'indicative')
INDF = SmartTag('INDF', 'indefinite')
INF = SmartTag('INF', 'infinitive')
INS = SmartTag('INS', 'instrumental')
INTR = SmartTag('INTR', 'intransitive')
IPFV = SmartTag('IPFV', 'imperfective')
IRR = SmartTag('IRR', 'irrealis')
LOC = SmartTag('LOC', 'locative')
M = SmartTag('M', 'masculine')
N = SmartTag('N', 'neuter')
NON = SmartTag('N-', 'non- (e.g. NSG nonsingular, NPST nonpast)')
NEG = SmartTag('NEG', 'negation, negative')
NMLZ = SmartTag('NMLZ', 'nominalizer/nominalization')
NOM = SmartTag('NOM', 'nominative')
OBJ = SmartTag('OBJ', 'object')
OBL = SmartTag('OBL', 'oblique')
P = SmartTag('P', 'patient-like argument of canonical transitive verb')
PASS = SmartTag('PASS', 'passive')
PFV = SmartTag('PFV', 'perfective')
PL = SmartTag('PL', 'plural')
POSS = SmartTag('POSS', 'possessive')
PRED = SmartTag('PRED', 'predicative')
PRF = SmartTag('PRF', 'perfect')
PRS = SmartTag('PRS', 'present')
PROG = SmartTag('PROG', 'progressive')
PROH = SmartTag('PROH', 'prohibitive')
PROX = SmartTag('PROX', 'proximal/proximate')
PST = SmartTag('PST', 'past')
PTCP = SmartTag('PTCP', 'participle')
PURP = SmartTag('PURP', 'purposive')
Q = SmartTag('Q', 'question particle/marker')
QUOT = SmartTag('QUOT', 'quotative')
RECP = SmartTag('RECP', 'reciprocal')
REFL = SmartTag('REFL', 'reflexive')
REL = SmartTag('REL', 'relative')
RES = SmartTag('RES', 'resultative')
S = SmartTag('S', 'single argument of canonical intransitive verb')
SBJ = SmartTag('SBJ', 'subject')
SBJV = SmartTag('SBJV', 'subjunctive')
SG = SmartTag('SG', 'singular')
TOP = SmartTag('TOP', 'topic')
TR = SmartTag('TR', 'transitive')
VOC = SmartTag('VOC', 'vocative')



# -*- coding: utf-8 -*-

'''
Data mapping functions
Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import logging
import threading
import json
from json import JSONDecoder, JSONEncoder


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------

class IDGenerator(object):

    def __init__(self, id_seed=1, id_hook=None):
        ''' id_seed = starting number '''
        self.__id_seed = id_seed
        self.__id_check_hook = id_hook  # external ID checker
        self.__lock = threading.Lock()

    def __next__(self):
        with self.__lock:
            while True:
                valid_id = self.__id_seed
                self.__id_seed += 1
                if self.__id_check_hook is None or not self.__id_check_hook(valid_id):
                    break
            return valid_id


class DataObject(object):

    def __init__(self, **kwargs):
        self.__extra_data = {}  # for internal purpose
        add_extra_fields(self, kwargs)

    def __getattr__(self, attr_name):
        return self.__extra_data[attr_name] if attr_name in self.__extra_data else None

    def update(self, a_dict, *fields, **field_map):
        flex_update_obj(a_dict, self, True, *fields, **field_map)

    def to_json(self, *args, **kwargs):
        return dumps(self, *args, **kwargs)

    def to_dict(self, *args, **kwargs):
        a_dict = to_json(self)
        if '_DataObject__extra_data' in a_dict and not a_dict['_DataObject__extra_data']:
            a_dict.pop('_DataObject__extra_data')
        return a_dict


def field(f, field_map):
    return f if f not in field_map else field_map[f]


def add_extra_fields(obj, kwargs):
    for k, v in kwargs.items():
        if k not in obj.__dict__:
            obj.__dict__[k] = v
        else:
            raise Exception("Attribute {} exists".format(k))


def dumps(obj, *args, **kwargs):
    ''' Typeless dump an object to json string '''
    return json.dumps(obj, *args, cls=TypelessSONEncoder, ensure_ascii=False, **kwargs)


def update_obj(source, target, *fields, **field_map):
    flex_update_obj(source, target, False, *fields, **field_map)


def flex_update_obj(source, target, __silent, *fields, **field_map):
    ''' Pull data from source to target.
    Target's __dict__ (object data) will be used by default. Otherwise, it'll be treated as a dictionary '''
    source_dict = source.__dict__ if hasattr(source, '__dict__') else source
    if not fields:
        fields = source_dict.keys()
    for f in fields:
        if f not in source_dict and __silent:
            continue
        target_f = f if f not in field_map else field_map[f]
        setattr(target, target_f, source_dict[f])


def to_json(obj, *fields, **field_map):
    if isinstance(obj, set):
        return list(obj)
    obj_dict = obj.__dict__ if hasattr(obj, '__dict__') else obj
    if not fields:
        fields = obj_dict.keys()
    json_dict = {field(k, field_map): obj_dict[k] for k in fields}
    return json_dict


def to_obj(cls, obj_data=None, *fields, **field_map):
    ''' Use obj_data (dict-like) to construct an object of type cls
    prioritize obj_dict when there are conficts '''
    if not fields:
        fields = obj_data.keys()
    try:
        kwargs = {field(f, field_map): obj_data[f] for f in fields if f in obj_data}
        obj = cls(**kwargs)
    except:
        getLogger().exception("Couldn't use kwargs to construct object")
        # use default constructor
        obj = cls()
        update_obj(obj_data, obj, *fields, **field_map)
    return obj


class TypedJSONEncoder(JSONEncoder):

    def __init__(self, *args, type_map=None, **kwargs):
        JSONEncoder.__init__(self, *args, **kwargs)
        self.type_map = type_map if type_map else {}

    def default(self, obj):
        if obj.__class__ in self.type_map:
            nj = to_json(obj)
            nj['__type__'] = self.type_map[obj.__class__]
            return nj
        else:
            return JSONEncoder.default(self, obj)


class TypelessSONEncoder(JSONEncoder):

    def __init__(self, *args, type_map=None, **kwargs):
        JSONEncoder.__init__(self, *args, **kwargs)
        self.type_map = type_map if type_map else {}

    def default(self, obj):
        try:
            return JSONEncoder.default(self, obj)
        except TypeError:
            # fall back to to_json helper function
            return to_json(obj)


class TypedJSONDecoder(JSONDecoder):

    def __init__(self, type_map=None, **kwargs):
            JSONDecoder.__init__(self, object_hook=self.obj_hooker)
            self.type_map = kwargs
            if type_map:
                self.type_map.update(type_map)

    def obj_hooker(self, obj_dict):
        if '__type__' in obj_dict:
            # check if this type is supported
            if obj_dict['__type__'] in self.type_map:
                obj_type = obj_dict.pop('__type__')
                return to_obj(self.type_map[obj_type], obj_dict)
            else:
                getLogger().warning("Unknown type: {}".format(obj_dict['__type__']))
        else:
            return obj_dict



# -*- coding: utf-8 -*-

''' Chirptext Test Scripts
Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

# This source code is a part of chirptext library
# Copyright (c) 2015, Le Tuan Anh <tuananh.ke@gmail.com>
# LICENSE: The MIT License (MIT)
#
# Homepage: https://github.com/letuananh/chirptext

import os
from chirptext.cli import setup_logging

TEST_DIR = os.path.dirname(__file__)
TEST_DATA = os.path.join(TEST_DIR, 'data')

setup_logging(os.path.join(TEST_DIR, 'logging.json'), os.path.join(TEST_DIR, 'logs'))



# -*- coding: utf-8 -*-

'''
Dao Phay library: A collection of tools for processing Vietnamese text using Python.

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

# in lower case: aàảãáạăằẳẵắặâầẩẫấậbcdđeèẻẽéẹêềểễếệghiìỉĩíịklmnoòỏõóọôồổỗốộơờởỡớợpqrstuùủũúụưừửữứựvxyỳỷỹýỵ
VIETNAMESE_ALPHABET_ORDER = 'aAàÀảẢãÃáÁạẠăĂằẰẳẲẵẴắẮặẶâÂầẦẩẨẫẪấẤậẬbBcCdDđĐeEèÈẻẺẽẼéÉẹẸêÊềỀểỂễỄếẾệỆgGhHiIìÌỉỈĩĨíÍịỊkKlLmMnNoOòÒỏỎõÕóÓọỌôÔồỒổỔỗỖốỐộỘơƠờỜởỞỡỠớỚợỢpPqQrRsStTuUùÙủỦũŨúÚụỤưƯừỪửỬữỮứỨựỰvVxXyYỳỲỷỶỹỸýÝỵỴ'
VIETNAMESE_SORTING_DICT = dict((x, VIETNAMESE_ALPHABET_ORDER.index(x)) for x in VIETNAMESE_ALPHABET_ORDER)
python_sorted = sorted


# ------------------------------------------------------------------------------
# Functions
# ------------------------------------------------------------------------------

def vnorder_char(c):
    return VIETNAMESE_SORTING_DICT[c] if c in VIETNAMESE_SORTING_DICT else ord(c)


def vnorder(s):
    return [vnorder_char(c) for c in s]


def sorted(list_of_strings):
    return python_sorted(list_of_strings, key=vnorder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Script for testing leutile

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import io
import unittest
import logging
import json
from chirptext import TextReport
from chirptext import texttaglib as ttl
from chirptext.deko import parse

# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(__file__)
TEST_DATA = os.path.join(TEST_DIR, 'data')
TEST_FILE = os.path.join(TEST_DATA, 'test')

BARK_SID = '01047745-v'
GDOG_SID = '02103841-n'


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Tests
# ------------------------------------------------------------------------------

class TestBasicModel(unittest.TestCase):

    def test_tag_model(self):
        ssid = '06162979-n'
        tag = ttl.Tag(ssid, tagtype='PWN30')
        self.assertEqual(tag.type, 'PWN30')
        self.assertEqual(tag.tagtype, 'PWN30')
        self.assertEqual(tag.text, ssid)
        self.assertEqual(tag.label, ssid)
        tag.text = '00636921-n'
        self.assertEqual(tag.label, '00636921-n')
        tag.type = 'OMW1.2'
        self.assertEqual(tag.tagtype, 'OMW1.2')
        self.assertEqual(tag.to_json(), {'label': '00636921-n', 'type': 'OMW1.2'})
        # no type
        tag_foo = ttl.Tag('foo')
        tag_foo.source = '頭'
        self.assertEqual(repr(tag_foo), '`foo`')
        self.assertEqual(tag_foo.to_json(), {'label': 'foo', 'source': '頭'})
        # from_json
        tag_json = {'label': 'foo', 'source': '頭', 'type': '冗談', 'cfrom': 0, 'cto': 3}
        tag_new = ttl.Tag.from_json(tag_json)
        self.assertEqual(tag_json, tag_new.to_json())

    def test_sentid(self):
        doc = ttl.Document('mydoc')
        sent = doc.new_sent('First sentence.')
        self.assertEqual(sent.ID, 1)
        sent2 = doc.new_sent('Second sentence.')
        self.assertEqual(sent2.ID, 2)
        # add some sentences manually
        sentm1 = ttl.Sentence('Another one', ID=3)
        sentm2 = ttl.Sentence('Another one 2', ID='5')
        doc.add_sent(sentm1)
        doc.add_sent(sentm2)
        doc.new_sent('Third sentence.')
        doc.new_sent('Fourth sentence.')
        sent5 = doc.new_sent('Fifth sentence.')
        self.assertEqual(sent5.ID, 7)
        # cannot add 3 again
        sent_foo = ttl.Sentence('Foo sentence.', ID=3)
        self.assertRaises(Exception, lambda: doc.add_sent(sent_foo))
        # cannot add a None sentence
        self.assertRaises(Exception, lambda: doc.add_sent(None))
        # document should have 5 created sentences + 2 imported sentences
        self.assertEqual(len(doc), 7)

    def test_extra_fields(self):
        cmt = 'This sentence is in English'
        s = ttl.Sentence(text='I am a sentence.', docID=1, comment=cmt)
        self.assertEqual(s.docID, 1)
        self.assertEqual(s.comment, cmt)

    def test_creating_sent(self):
        tokens = 'It rains .'.split()
        s = ttl.Sentence('It rains.', tokens=tokens)
        self.assertEqual([t.text for t in s.tokens], tokens)
        self.assertEqual(repr(s), s.text)
        self.assertEqual(str(s), s.text)
        self.assertEqual(s.surface(s[0]), 'It')
        self.assertEqual(s.surface(ttl.Tag(cfrom=-1, cto=-1)), '')
        self.assertEqual(s.surface(ttl.Tag(cfrom=-1, cto=3)), '')
        self.assertEqual(s.surface(ttl.Tag(cfrom=3, cto=-1)), '')
        self.assertEqual(s.surface(ttl.Tag(cfrom=None, cto=None)), '')
        s.ID = '1'
        self.assertEqual(repr(s), '#1: It rains.')
        self.assertEqual(str(s), '#1: It rains.')
        # tag sentence
        url = 'https://github.com/letuananh/chirptext'
        turl = s.new_tag(url, tagtype='URL')
        self.assertEqual(s.get_tag('URL').text, url)
        self.assertEqual(s.tagmap(), {'URL': [turl]})
        self.assertEqual(s.get_tags('URL'), [turl])
        # test concepts
        c = ttl.Concept(tag='02756558-v', clemma='rain')
        s.add_concept(c)
        self.assertRaises(Exception, lambda: s.add_concept(None))
        self.assertRaises(Exception, lambda: s.add_concept(c))
        c2 = ttl.Concept(tag='dummy', clemma='it')
        c2.cidx = c.cidx
        self.assertRaises(Exception, lambda: s.add_concept(c2))
        c2.cidx = s.new_concept_id()
        s.add_concept(c2)
        self.assertEqual(len(s.concepts), 2)
        self.assertRaises(Exception, lambda: s.pop_concept(3))
        self.assertEqual(s.pop_concept(3, default=c2), c2)
        self.assertEqual(s.pop_concept(c2.cidx), c2)
        self.assertEqual(s.concepts, [c])

    def test_idgen(self):
        idg = ttl.IDGenerator()
        _ids = [next(idg) for i in range(3)]
        idg_explicit = ttl.IDGenerator(id_seed=1)
        _ids2 = [next(idg_explicit) for i in range(3)]
        expected = [1, 2, 3]
        self.assertEqual(_ids, expected)
        self.assertEqual(_ids2, expected)


class TestBuildTags(unittest.TestCase):

    def test_create_sent(self):
        doc = ttl.Document('test', TEST_DATA)
        # add words
        sent = doc.new_sent('Some happy guard-dogs barked.', 1)
        sent.import_tokens('Some happy guard dogs barked .'.split())
        self.assertEqual(len(sent), 6)
        # sense tagging
        sent.new_concept('01148283-a', 'happy', tokens=[sent[1]])
        c = sent.new_concept('02084071-n', 'dog')
        sent.concept(c.cidx).add_token(sent[3])
        sent.new_concept(BARK_SID, 'bark').add_token(sent[4])
        sent.new_concept(GDOG_SID, 'guard-dog').add_token(sent[2], sent[3])  # MWE example
        # verification
        tcmap = sent.tcmap()
        self.assertEqual(tcmap[sent[3]][0].clemma, 'dog')
        self.assertEqual(tcmap[sent[3]][1].clemma, 'guard-dog')
        self.assertEqual(tcmap[sent[3]][1].tag, GDOG_SID)
        self.assertEqual(tcmap[sent[4]][0].tag, BARK_SID)
        mwe = list(sent.mwe())
        self.assertTrue(mwe)
        self.assertEqual(mwe[0].tag, GDOG_SID)

    def test_sentids(self):
        doc = ttl.Document('boo')
        doc.add_sent(ttl.Sentence('odd', ID=3))  # add sent#3 first
        doc.new_sent('foo')  # 1
        doc.new_sent('boo')  # 2
        moo = doc.new_sent('moo')  # moo will be #4 because sent #3 exists
        self.assertEqual(moo.ID, 4)
        sids = [s.ID for s in doc]
        self.assertEqual(sids, [3, 1, 2, 4])  # sidx is not the same as sent.ID


class TestComment(unittest.TestCase):
    ''' Ensure that all objects may hold comments and flags '''

    def test_storing_flags(self):
        ''' '''
        doc = ttl.read(TEST_FILE)
        sent = doc[0]
        sent.flag = ttl.Tag.GOLD
        sent[0].flag = ttl.Tag.NLTK
        sent.concepts[0].flag = ttl.Tag.ISF
        sent_json = sent.to_json()
        self.assertEqual(sent_json['flag'], ttl.Tag.GOLD)
        self.assertEqual(sent_json['tokens'][0]['flag'], ttl.Tag.NLTK)
        self.assertEqual(sent_json['concepts'][0]['flag'], ttl.Tag.ISF)

class TestTagging(unittest.TestCase):

    def test_taginfo(self):
        print("Test tag info")
        t = ttl.Tag('dog', 1, 4, source=ttl.Tag.ISF)
        self.assertEqual(t.cfrom, 1)
        self.assertEqual(t.cto, 4)
        self.assertEqual(t.label, 'dog')
        self.assertEqual(t.source, ttl.Tag.ISF)

        # Source = DEFAULT now
        t2 = ttl.Tag('bark', 0, 5)
        self.assertEqual(t2.cfrom, 0)
        self.assertEqual(t2.cto, 5)
        self.assertEqual(t2.label, 'bark')
        self.assertEqual(t2.source, ttl.Tag.NONE)

    def test_tag_token(self):
        token = ttl.Token("Words", 0, 4)
        token.new_tag("plural")
        token.new_tag("word-token")
        token.pos = 'n'
        token.lemma = 'word'
        token.comment = "an element of speech or writing"
        js_token = token.to_json()
        getLogger().debug(js_token)
        expected = {'cfrom': 0, 'cto': 4, 'text': 'Words', 'lemma': 'word', 'pos': 'n', 'comment': 'an element of speech or writing', 'tags': [{'label': 'plural'}, {'label': 'word-token'}]}
        self.assertEqual(js_token, expected)

    def test_tag_type_and_searching(self):
        taggable_objects = [
            ttl.Token("text", 0, 4),
            ttl.Sentence('I am a sentence.')
        ]
        for obj in taggable_objects:
            obj.new_tag("06387980-n", tagtype="synset")
            obj.new_tag("06414372-n", tagtype="synset")
            obj.new_tag("manual", tagtype="tagtype")
            # find all tags by type
            synsets = [x.text for x in obj.get_tags('synset')]
            self.assertEqual(synsets, ["06387980-n", "06414372-n"])
            # find a specific tag
            self.assertEqual(obj.get_tag('tagtype').text, "manual")
            # not found ...
            self.assertEqual(obj.get_tag('meaning', default='N/A').text, "N/A")
            self.assertRaises(LookupError, lambda: obj.get_tag('meaning'))
            # auto create ..
            self.assertEqual(obj.get_tag('auto', auto_create=True).text, "")
            self.assertEqual(obj.get_tag('auto2', auto_create=True, default='X').text, "X")
            # the tags must persist
            self.assertEqual(obj.get_tag('auto').text, "")
            self.assertEqual(obj.get_tag('auto2').text, "X")
            # test deprecation warning
            if isinstance(obj, ttl.Token):
                with self.assertWarnsRegex(Warning, 'Token.find\(\) is deprecated and will be removed in near future. Use Token.get_tag\(\) instead'):
                    obj.find("auto")
                with self.assertWarnsRegex(Warning, 'Token.find_all\(\) is deprecated and will be removed in near future. Use Token.get_tags\(\) instead'):
                    obj.find_all("auto2")

    def import_tokens(self, sent, token_list):
        sent.tokens = token_list

    def test_import_tokens(self):
        sent = ttl.Sentence('It rains.')
        tokens = ['It', 'rains', '.']
        sent.tokens = tokens
        self.assertEqual([t.text for t in sent.tokens], tokens)
        # cannot import twice
        self.assertRaises(Exception, lambda: self.import_tokens(sent, tokens))
        # or import half-way
        sent2 = ttl.Sentence("Cats don't like cats that meow.")
        sent2.import_tokens(('Cats',))
        tokens2 = "do n't like cats that meow .".split()
        self.assertRaises(Exception, lambda: self.import_tokens(sent, tokens2))
        sent2.import_tokens(tokens2)  # but use import_tokens explicitly is fine
        self.assertEqual([t.text for t in sent2.tokens], ['Cats', 'do', "n't", 'like', 'cats', 'that', 'meow', '.'])

    def test_comment(self):
        sent = ttl.Sentence("Dogs bark.")
        sent.import_tokens("Dogs bark .".split())
        sent.comment = 'I am a test sentence.'
        sent[0].comment = "canine"
        sent.new_concept("02084071-n", "dog", tokens=(sent[0],))
        sent.concepts[0].comment = 'a member of the genus Canis (probably descended from the common wolf) that has been domesticated by man since prehistoric times'
        expected = {'text': 'Dogs bark.', 'comment': 'I am a test sentence.', 'tokens': [{'cto': 4, 'cfrom': 0, 'comment': 'canine', 'text': 'Dogs'}, {'cto': 9, 'cfrom': 5, 'text': 'bark'}, {'cto': 10, 'cfrom': 9, 'text': '.'}], 'concepts': [{'tag': '02084071-n', 'clemma': 'dog', 'comment': 'a member of the genus Canis (probably descended from the common wolf) that has been domesticated by man since prehistoric times', 'tokens': [0]}]}
        getLogger().debug(sent.to_json())
        getLogger().debug(expected)
        self.assertEqual(expected, sent.to_json())
        self.assertFalse(sent.tags)
        sent.new_tag(GDOG_SID, 0, 4, tagtype='wn30')
        sent.new_tag(BARK_SID, 5, 9, tagtype='wn30')
        for t in sent.tags:
            getLogger().debug("{}: label={} | type = {}".format(t, t.label, t.tagtype))

    def test_tagged_sentences(self):
        print("test converting MeCabSent into TTL Sent manually")
        sent = ttl.Sentence('猫が好きです 。')
        mecab_sent = parse(sent.text)
        getLogger().debug((mecab_sent.surface, mecab_sent.words))
        # import tags
        sent.import_tokens(mecab_sent.words)
        for mtoken, token in zip(mecab_sent, sent.tokens):
            if mtoken.reading_hira() != token.text:
                if token.text in ('猫', '好き'):
                    token.new_tag(mtoken.reading_hira(), tagtype='reading')
                token.lemma = mtoken.reading_hira()
                token.pos = mtoken.pos3()
        self.assertEqual(mecab_sent.words, [x.text for x in sent.tokens])
        self.assertEqual(sent[0].get_tag('reading').label, 'ねこ')
        self.assertEqual(sent[0].lemma, 'ねこ')
        self.assertEqual(sent[2][0].label, 'すき')
        self.assertFalse(sent[3].lemma, '')  # if there is no lemma label => return ''
        self.assertEqual(sent[2].surface(), '好き')
        self.assertFalse(len(sent[1]))
        self.assertFalse(len(sent[3]))
        self.assertFalse(len(sent[4]))
        return sent

    def test_tagged_sent_to_json(self):
        sent = ttl.Sentence("女の子は猫が好きです。")
        sent.import_tokens("女 の 子 は 猫 が 好き です 。".split())
        sent[0].lemma = "おんな"
        sent[2].lemma = "こ"
        sent[4].lemma = "ねこ"
        sent[4].comment = "Say neh-koh"
        sent[4].pos = "名詞-一般"
        sent[6].lemma = "すき"
        sent[6].pos = "名詞-形容動詞語幹"
        sent.new_concept("10084295-n", "女の子", (sent[0], sent[1], sent[2]))
        sent.concepts[0].comment = "若々しい女の人"
        sent.new_concept("02121620-n", clemma="猫").add_token(sent[4])
        sent.new_concept("01292683-a", clemma="好き").add_token(sent[6])
        expected = {'tokens': [{'cfrom': 0, 'cto': 1, 'lemma': 'おんな', 'text': '女'},
                               {'cfrom': 1, 'cto': 2, 'text': 'の'},
                               {'cfrom': 2, 'cto': 3, 'lemma': 'こ', 'text': '子'},
                               {'cfrom': 3, 'cto': 4, 'text': 'は'},
                               {'pos': '名詞-一般', 'cfrom': 4, 'cto': 5, 'lemma': 'ねこ', 'text': '猫', 'comment': 'Say neh-koh'},
                               {'cfrom': 5, 'cto': 6, 'text': 'が'},
                               {'pos': '名詞-形容動詞語幹', 'cfrom': 6, 'cto': 8, 'lemma': 'すき', 'text': '好き'},
                               {'cfrom': 8, 'cto': 10, 'text': 'です'},
                               {'cfrom': 10, 'cto': 11, 'text': '。'}],
                    'text': '女の子は猫が好きです。',
                    'concepts': [{'tag': '10084295-n', 'tokens': [0, 1, 2], 'clemma': '女の子', 'comment': '若々しい女の人'},
                                 {'tag': '02121620-n', 'tokens': [4], 'clemma': '猫'},
                                 {'tag': '01292683-a', 'tokens': [6], 'clemma': '好き'}]}
        actual = sent.to_json()
        self.assertEqual(expected['text'], actual['text'])
        self.assertEqual(expected['concepts'], actual['concepts'])
        self.assertEqual(expected['tokens'], actual['tokens'])
        self.assertEqual(expected, actual)
        getLogger().debug(actual)

    def test_json_tagged_sent(self):
        raw = {'text': '女の子は猫が好きです。', 'tokens': [{'cfrom': 0, 'cto': 1, 'text': '女', 'lemma': 'おんな'}, {'cfrom': 1, 'cto': 2, 'text': 'の'}, {'cfrom': 2, 'cto': 3, 'text': '子', 'lemma': 'こ'}, {'cfrom': 3, 'cto': 4, 'text': 'は'}, {'cfrom': 4, 'cto': 5, 'text': '猫', 'lemma': 'ねこ', 'pos': '名詞-一般', 'comment': 'Say neh-koh'}, {'cfrom': 5, 'cto': 6, 'text': 'が'}, {'cfrom': 6, 'cto': 8, 'text': '好き', 'lemma': 'すき', 'pos': '名詞-形容動詞語幹'}, {'cfrom': 8, 'cto': 10, 'text': 'です'}, {'cfrom': 10, 'cto': 11, 'text': '。'}], 'concepts': [{'clemma': '女の子', 'tag': '10084295-n', 'tokens': [0, 1, 2], 'comment': '若々しい女の人', 'flag': 'G'}, {'clemma': '猫', 'tag': '02121620-n', 'tokens': [4]}, {'clemma': '好き', 'tag': '01292683-a', 'tokens': [6]}]}
        # json => TaggedSentence => json
        sent = ttl.Sentence.from_json(raw)
        sent_json = sent.to_json()
        self.assertEqual(raw['text'], sent_json['text'])
        self.assertEqual(raw['concepts'], sent_json['concepts'])
        self.assertEqual(raw['tokens'], sent_json['tokens'])
        self.assertEqual(raw, sent_json)
        self.assertEqual(sent_json['concepts'][0]['flag'], 'G')

    def test_tagging_erg_sent(self):
        ''' Test import tokens '''
        txt = '''In this way I am no doubt indirectly responsible for Dr. Grimesby Roylott's death, and I cannot say that it is likely to weigh very heavily upon my conscience."'''
        words = ['in', 'this', 'way', 'i', 'am', 'no', 'doubt', 'indirectly', 'responsible', 'for', 'dr.', 'Grimesby', 'Roylott', "'s", 'death', ',', 'and', 'i', 'can', 'not', 'say', 'that', 'it', 'is', 'likely', 'to', 'weigh', 'very', 'heavily', 'upon', 'my', 'conscience', '.', '"']
        s = ttl.Sentence(txt)
        s.import_tokens(words)
        self.assertEqual(words, [x.text for x in s.tokens])

    def test_tagged_doc(self):
        doc = ttl.read(TEST_FILE)
        ts_count = [(len(s), len(s.concepts)) for s in doc]
        self.assertEqual(len(doc), 3)  # 3 sents
        self.assertEqual(ts_count, [(29, 15), (28, 13), (34, 23)])

    def test_cwl(self):
        doc = ttl.read(TEST_FILE)
        sent = doc[0]
        # ensure that words and concepts are linked properly
        w = doc[0][-4]
        tcmap = sent.tcmap()
        c = tcmap[w][0]
        self.assertEqual(w.pos, 'WP')
        self.assertEqual(c.tokens[0].pos, 'WP')
        # test sentence level tagging
        self.assertTrue([str(t) for t in sent.tags])

    def test_multiple_tags(self):
        doc = ttl.read(TEST_FILE)
        mw_ms = [(len(list(s.mwe())), len(list(s.msw()))) for s in doc]
        self.assertEqual(mw_ms, [(2, 2), (0, 0), (3, 2)])

    def test_recover_surface_string(self):
        s = ttl.Sentence('''a religious sect founded in the United States in 1966; based on Vedic scriptures; groups engage in joyful chanting of `Hare Krishna' and other mantras based on the name of the Hindu god Krishna; devotees usually wear saffron robes and practice vegetarianism and celibacy''')
        tokens = ['a', 'religious', 'sect', 'founded', 'in', 'the', 'United', 'States', 'in', '1966', ';', 'based', 'on', 'Vedic', 'scriptures', ';', 'groups', 'engage', 'in', 'joyful', 'chanting', 'of', 'Hare', 'Krishna', 'and', 'other', 'mantras', 'based', 'on', 'the', 'name', 'of', 'the', 'Hindu', 'god', 'Krishna', ';', 'devotees', 'usually', 'wear', 'saffron', 'robes', 'and', 'practice', 'vegetarianism', 'and', 'celibacy']
        s.import_tokens(tokens)
        cfrom = min(x.cfrom for x in s.tokens)
        cto = max(x.cto for x in s.tokens)
        self.assertEqual(s.text, s.text[cfrom:cto])

    def test_export_to_streams(self):
        doc = ttl.Document('manual', TEST_DATA)
        # create sents in doc
        raws = ("三毛猫が好きです。", "雨が降る。", "女の子はケーキを食べる。")
        for sid, r in enumerate(raws):
            msent = parse(r)
            tsent = doc.new_sent(msent.surface, sid + 1)  # sentID starts from 1
            tsent.import_tokens(msent.words)
            # pos tagging
            for mtk, tk in zip(msent, tsent):
                tk.pos = mtk.pos3()
                tk.new_tag(mtk.reading_hira(), tagtype="Reading", source=ttl.Tag.MECAB)
        # sense tagging
        doc[2][4].comment = 'to eat'
        doc[0].new_concept("三毛猫", "wiki.ja:三毛猫", tokens=[0, 1, 2]).comment = 'Calico cat, you know?'
        doc[1].new_concept("降る", "02756821-v", tokens=(2,))
        doc[2].new_concept("女の子", "10084295-n", tokens=(0,))
        doc[2].new_concept("食べる", "01166351-v", (4,))
        # tags
        doc[0].new_tag("WIKI", 0, 3, tagtype="SRC")
        doc[0].new_tag("https://ja.wikipedia.org/wiki/三毛猫", 0, 3, tagtype="URL")
        doc[2].new_tag("WIKI", 0, 3, tagtype="SRC")
        doc[2].new_tag("https://ja.wikipedia.org/wiki/少女", 0, 3, tagtype="URL")
        # export doc
        concepts = TextReport.string()
        links = TextReport.string()
        sents = TextReport.string()
        tags = TextReport.string()
        words = TextReport.string()
        with ttl.TxtWriter(sents.file, words.file, concepts.file, links.file, tags.file) as writer:
            writer.write_doc(doc)
            getLogger().debug("sents\n{}".format(sents.content()))
            getLogger().debug("words\n{}".format(words.content()))
            getLogger().debug("concepts\n{}".format(concepts.content()))
            getLogger().debug("links\n{}".format(links.content()))
            getLogger().debug("tags\n{}".format(tags.content()))
            self.assertTrue(sents.content())
            self.assertTrue(words.content())
            self.assertTrue(concepts.content())
            self.assertTrue(links.content())
            self.assertTrue(tags.content())
            for sent in doc:
                logging.debug(json.dumps(sent.to_json(), ensure_ascii=False))


class TestSerialization(unittest.TestCase):

    def build_test_sent(self):
        sent = ttl.Sentence('三毛猫が好きです。')
        sent.flag = '0'
        sent.comment = 'written in Japanese'
        sent.new_tag('I like calico cats.', tagtype='eng')
        sent.import_tokens('三 毛 猫 が 好き です 。'.split())
        for tk, pos in zip(sent, '名詞 名詞 名詞 助詞 名詞 助動詞 記号'.split()):
            tk.pos = pos
        sent.new_concept("三毛猫", "wiki.ja:三毛猫", tokens=[0, 1, 2])
        sent[0].new_tag('mi', tagtype='reading')
        sent[1].new_tag('ke', tagtype='reading')
        sent[2].new_tag('neko', tagtype='reading')
        getLogger().debug(sent.to_json())
        return sent

    def test_json_serialization(self):
        print("Test serialization to and from JSON")
        sent = self.build_test_sent()
        sentj = sent.to_json()
        sent_re = ttl.Sentence.from_json(sentj)
        getLogger().debug(sent_re.to_json())
        self.assertEqual(sent_re.to_json(), sentj)

    def test_ttl_tsv_serialization(self):
        sent = self.build_test_sent()
        concepts = TextReport.string()
        links = TextReport.string()
        sents = TextReport.string()
        tags = TextReport.string()
        words = TextReport.string()
        writer = ttl.TxtWriter(sents.file, words.file, concepts.file, links.file, tags.file)
        writer.write_sent(sent)
        sents_txt = sents.content()
        words_txt = words.content()
        concepts_txt = concepts.content()
        links_txt = links.content()
        tags_txt = tags.content()
        getLogger().debug("sents\n{}".format(sents_txt))
        getLogger().debug("words\n{}".format(words_txt))
        getLogger().debug("concepts\n{}".format(concepts_txt))
        getLogger().debug("links\n{}".format(links_txt))
        getLogger().debug("tags\n{}".format(tags_txt))
        # read it back
        reader = ttl.TxtReader(io.StringIO(sents_txt),
                               io.StringIO(words_txt),
                               io.StringIO(concepts_txt),
                               io.StringIO(links_txt),
                               io.StringIO(tags_txt))
        docx = reader.read()
        # patch sent.ID
        sent.ID = 1
        jo = sent.to_json()
        jr = docx[0].to_json()
        getLogger().debug(jo)
        getLogger().debug(jr)
        self.assertEqual(jo['text'], jr['text'])
        self.assertEqual(jo['tokens'], jr['tokens'])
        self.assertEqual(jo['concepts'], jr['concepts'])
        self.assertEqual(jo['tags'], jr['tags'])
        self.assertEqual(jo['flag'], jr['flag'])
        self.assertEqual(jo['comment'], jr['comment'])
        self.assertEqual(jo, jr)


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test script for readers

Latest version can be found at https://github.com/letuananh/chirptext

:copyright: (c) 2012 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import gzip
import logging
import unittest
import json
from pathlib import Path
from chirptext.anhxa import to_json, to_obj
from chirptext import chio

# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_DATA = os.path.join(TEST_DIR, 'data')
TEST_CSV = os.path.join(TEST_DATA, 'test.csv')


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Data Structures
# -------------------------------------------------------------------------------

class Person:

    def __init__(self, first='', last='', idno=''):
        self.first = first
        self.last = last
        self.idno = idno

    @staticmethod
    def parse(full_name):
        parts = full_name.split()
        if len(parts) == 1:
            return Person(parts[0])
        elif len(parts) == 2:
            return Person(parts[0], parts[1])
        else:
            return Person(parts[0], parts[1], parts[2])

    def __repr__(self):
        return "Person(first={}, last={}, idno={})".format(repr(self.first), repr(self.last), repr(self.idno))

    def __eq__(self, other):
        return other and type(other) == Person and self.first == other.first and self.last == other.last and self.idno == other.idno


# -------------------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------------------


class TestIO(unittest.TestCase):

    def test_io_with_pathlib(self):
        print("Make sure that io functions works with pathlib.Path")
        # test read & write TXT
        data = [['name', 'foo'], ['age', '18']]
        json_path = Path(TEST_DATA) / 'temp.json'
        chio.write_file(json_path, json.dumps(data))
        json_data = json.loads(chio.read_file(json_path))
        self.assertEqual(json_data, data)
        # test read & write CSV
        filepath = Path(TEST_DATA) / 'temp.csv'
        chio.write_tsv(filepath, data)
        actual = chio.read_tsv(filepath)
        self.assertEqual(actual, data)

    def test_file_rw(self):
        tmpfile = os.path.join(TEST_DATA, 'test.txt')
        tmpgzfile = os.path.join(TEST_DATA, 'test.txt.gz')
        txt = 'ユニコード大丈夫だよ。'
        txtz = 'This is a zipped text file.'
        chio.write_file(content=txt, mode='wb', path=tmpfile)  # write content as bytes
        chio.write_file(tmpgzfile, content=txtz)
        # ensure that tmpgzfile is actually a gzip file
        with gzip.open(tmpgzfile, mode='rt') as infile:
            self.assertEqual(infile.read(), txtz)
        # verify written content
        self.assertTrue(chio.is_file(tmpfile))
        self.assertTrue(chio.is_file(tmpgzfile))
        self.assertEqual(chio.read_file(tmpfile), txt)
        self.assertEqual(chio.read_file(tmpgzfile), txtz)
        self.assertEqual(chio.read_file(tmpfile, mode='r'), txt)
        self.assertEqual(chio.read_file(tmpgzfile, mode='r'), txtz)
        self.assertIsInstance(chio.read_file(tmpfile, mode='rb'), bytes)
        self.assertIsInstance(chio.read_file(tmpgzfile, mode='rb'), bytes)


class TestUsingCSV(unittest.TestCase):

    def test_csv(self):
        data = [['abc', 'def\n"xyz'],
                ["'", '"', ',']]
        chio.write_csv(TEST_CSV, data)
        indata = chio.read_csv(TEST_CSV)
        self.assertEqual(data, indata)


class TestReaders(unittest.TestCase):

    def test_csv(self):
        print("Test CSV io")
        # test writing numbers => strings
        data = ((1, 2, 3), (4, 5, 6))
        chio.write_csv(TEST_CSV, data)
        indata = chio.read_csv(TEST_CSV)
        self.assertEqual(indata, [['1', '2', '3'], ['4', '5', '6']])

        # test writing strings with spaces
        data = [["Doraemon", "Nobita Nobi", "Shizuka Minamoto", "Dorami"],
                ["Takeshi Goda", "Suneo Honekawa", "Jaiko", "Hidetoshi Dekisugi"]]
        chio.write_tsv(TEST_CSV, data)
        indata = chio.read_csv(TEST_CSV)
        self.assertEqual(data, indata)

        # writing only 1 string with space?
        data = [['Takeshi Goda']]
        chio.write_tsv(TEST_CSV, data, quoting=None)
        indata = chio.read_csv(TEST_CSV)
        expected = [['Takeshi', 'Goda']]  # not quoting => CSV reader cannot guess the format
        self.assertEqual(indata, expected)
        # read file with a proper dialect
        indata = chio.read_csv(TEST_CSV, dialect='excel-tab')
        self.assertEqual(data, indata)

        # write file with forced quoting (default option)
        data = [['Takeshi Goda']]
        chio.write_tsv(TEST_CSV, data)
        indata = chio.read_csv(TEST_CSV)
        self.assertEqual(data, indata)

        # with header
        names = ["Doraemon - D001", "Nobita Nobi N001", "Shizuka Minamoto M001", "Dorami - D002",
                 "Takeshi Goda G001", "Suneo Honekawa H001", "Jaiko - G002", "Hidetoshi Dekisugi D003"]
        persons = [Person.parse(name) for name in names]
        header = ['first', 'last']
        chio.write_csv(TEST_CSV, [to_json(p) for p in persons], fieldnames=header)
        inrows = chio.read_csv(TEST_CSV, fieldnames=True)
        getLogger().debug("Inrows: {}".format(inrows))
        expected = [{'first': 'Doraemon', 'last': '-'}, {'first': 'Nobita', 'last': 'Nobi'}, {'first': 'Shizuka', 'last': 'Minamoto'}, {'first': 'Dorami', 'last': '-'}, {'first': 'Takeshi', 'last': 'Goda'}, {'first': 'Suneo', 'last': 'Honekawa'}, {'first': 'Jaiko', 'last': '-'}, {'first': 'Hidetoshi', 'last': 'Dekisugi'}]
        self.assertEqual(inrows, expected)
        # read in as objects
        csv_rows = chio.read_csv(TEST_CSV, fieldnames=True)
        inpersons = [to_obj(Person, row) for row in csv_rows]
        expected_names = [to_obj(Person, row) for row in expected]  # without idno
        self.assertEqual(inpersons, expected_names)

        # with idno
        header = ['first', 'last', 'idno']
        chio.write_csv(TEST_CSV, [to_json(p) for p in persons], fieldnames=header)
        inpersons = [to_obj(Person, row) for row in chio.read_csv(TEST_CSV, fieldnames=True)]
        self.assertEqual(persons, inpersons)


# -------------------------------------------------------------------------------
# Main
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()
